
dens.univ.os.int <- function(symbol, distribution="Normal", param, log=TRUE){
  
  # symbol should be a list containing:
  # 1) a vector "coord" of length 2 or 5, according to type
  # 2) a vector "orders" of length 2 or 5
  # 3) a integer "n" giving the number of points aggregated
  #
  # parameter vector should be of length 2: (mean, var)

  if(!is.list(symbol)){stop("Symbol should be a list.")}
  if(length(param) != 2){stop("Wrong length of parameter vector")}
  if( param[2]<=0){ if(log){ return( -1e300 ) }else{ return( 1e-300) }}
  if(is.null(symbol$n)){stop("Symbol should contain sample size n")}
  n <- symbol$n
  coord <- c(-Inf, symbol$coord, Inf)
  order <- c(0, symbol$order, n+1)
  B <- length(symbol$coord)+1 # Number of bins in "histogram"
  
  if(distribution=="Normal"){
    if(log){
      d <- sum(dnorm(symbol$coord, mean=param[1], sd=param[2], log=TRUE))
    }else{d <- prod(dnorm(symbol$coord, mean=param[1], sd=param[2])) }
    Probs <- pnorm(coord, mean=param[1], sd=param[2])
  }
  if(distribution=="l-Normal"){
    if(log){
      d <- sum(dlnorm(symbol$coord, meanlog=param[1], sdlog=param[2], log=TRUE))
    }else{d <- prod(dlnorm(symbol$coord, meanlog=param[1], sdlog=param[2])) }
    Probs <- plnorm(coord, meanlog=param[1], sdlog=param[2])
  }
  
  for(i in 1:B){
    prob <- Probs[i+1] - Probs[i]
    if(prob <= 0){prob <- 1e-320}
    if(log){
      d <- d + log( prob ) * (order[i+1]-order[i]-1)
    }else{
      d <- d * ( prob )^(order[i+1]-order[i]-1)
    }  
  }
  
  return(d)  

}

llik.univ.os.int <- function(symbol.data, distribution="Normal", param=c(0,1), log=TRUE, vectorial=FALSE){
  # symbol.data is the symbolic dataset, a list of symbols as defined in dens.biv.int
  # If vectorial=TRUE then a vector of each log-likelihood evaluation is returned, otherwise the sum is returned
  
  m <- length(symbol.data) # number of symbols
  llik <- vector("numeric",length=m)
  
  for( j in 1:m){
    llik[j] <- dens.univ.os.int(symbol=symbol.data[[j]], distribution=distribution, param=param, log=log)
  }
  
  if(vectorial){
    return(llik)
  }else{
    return(sum(llik))
  }
  
}  

# Optimisation of the log-likelihood function

fit.univ.os.int <- function(symbol.data, distribution="Normal", parastart, hessian=TRUE, trace=0){
  optimfun <- function(para){
    #cat("param:", para, "\n")
    return(llik.univ.os.int(symbol.data=symbol.data, distribution=distribution, param=para, log=TRUE))
  }
  est.para <- optim(parastart, optimfun, method="BFGS", control = list(maxit = 2000, trace = trace, fnscale=-1), hessian=hessian)
  if(hessian){
    return(list(est=est.para$par, stderr=sqrt(diag(solve(-est.para$hessian))), hess=est.para$hessian, LL=est.para$value))
  }else{
    return(list(est=est.para$par ))
  }
  
}
