### Likelihood function for bivariate intervals
### assuming that we have no extra knowledge than
### the upper and lower bounds
###
### Intervals are given by four points:
###   (a1,b1), (a1,b2), (a2,b1) and (a2,b2)
### where a1 < a2 and b1 < b2
###
### Making use of the function dens.biv.int which 
### considers extra information 
###
### We focus on the bivariate Normal distribution

library(mvtnorm)

dens.biv.int.full <- function(symbol, param=c(0,0,1,0,1), p.bltr, p.tlbr, p.tl, p.tr, p.br, p.bl, p.four, log=TRUE){

  # symbol should be a list containing:
  # 1) a vector "coord" of length 4 with a1, a2, b1, b2
  # 2) an integer giving the number of points in the interval
  #
  # parameter vector should be of length 5: (mean1, mean2, var1, cov, var2)
  #   var1 and var2 should be positive

  if(!is.list(symbol)){stop("Symbol should be a list.")}
  if( length(symbol$coord) !=4){stop(" 'symbol$coord' should be of lentgh 4.")}
  if(length(param) != 5){stop("Wrong length of parameter vector")}
  if( param[3]<=0 || param[5]<=0 ){
    if(log){
      return( -1e300 )
    }else{
      return( 1e-300)
    }
  }

  symbol[["extra"]] <- vector(length=2)
  n <- symbol$n

  # Likelihood BLTR times the probability of BLTR according to the parameters
  symbol$extra <- c(2, "bltr")
  pt.bltr <- n * (n-1) * dens.biv.int(symbol, param=param, log=FALSE) * p.bltr

  # Likelihood TLBR times the probability of TLBR according to the parameters
  symbol$extra <- c(2, "tlbr")
  pt.tlbr <- n * (n-1) * dens.biv.int(symbol, param=param, log=FALSE) * p.tlbr

  # Likelihood TL times the probability of TL according to the parameters
  symbol$extra <- c(3, "tl")
  pt.tl <- n * (n-1) * (n-2) * dens.biv.int(symbol, param=param, log=FALSE) * p.tl

  # Likelihood TR times the probability of TR according to the parameters
  symbol$extra <- c(3, "tr")
  pt.tr <- n * (n-1) * (n-2) * dens.biv.int(symbol, param=param, log=FALSE) * p.tr

  # Likelihood BR times the probability of BR according to the parameters
  symbol$extra <- c(3, "br")
  pt.br <- n * (n-1) * (n-2) * dens.biv.int(symbol, param=param, log=FALSE) * p.br

  # Likelihood BL times the probability of BL according to the parameters
  symbol$extra <- c(3, "bl")
  pt.bl <- n * (n-1) * (n-2) * dens.biv.int(symbol, param=param, log=FALSE) * p.bl

  # Likelihood 4 times the probability of 4 according to the parameters
  symbol$extra <- c(4, 0)
  pt.four <- n * (n-1) * (n-2) * (n-3) * dens.biv.int(symbol, param=param, log=FALSE) * p.four

  if(log){
    return( log(pt.bltr + pt.tlbr + pt.tl + pt.tr + pt.br + pt.bl + pt.four))
  }else{
    return( pt.bltr + pt.tlbr + pt.tl + pt.tr + pt.br + pt.bl + pt.four )
  }

}

# dens.biv.int.full <- function(symbol, param=c(0,0,1,0,1), p.bltr, p.tlbr, p.tl, p.tr, p.br, p.bl, p.four, log=TRUE){
# 
#   # symbol should be a list containing:
#   # 1) a vector "coord" of length 4 with a1, a2, b1, b2
#   # 2) an integer giving the number of points in the interval
#   #
#   # parameter vector should be of length 5: (mean1, mean2, var1, cov, var2)
#   #   var1 and var2 should be positive
# 
#   if(!is.list(symbol)){stop("Symbol should be a list.")}
#   if( length(symbol$coord) !=4){stop(" 'symbol$coord' should be of lentgh 4.")}
#   if(length(param) != 5){stop("Wrong length of parameter vector")}
#   if( param[3]<=0 || param[5]<=0 ){
#     if(log){
#       return( -1e300 )
#     }else{
#       return( 1e-300)
#     }
#   }
# 
#   symbol[["extra"]] <- vector(length=2)
# 
#   # Max <-  max(p.bltr, p.tlbr, p.tl, p.tr, p.br, p.bl, p.four)
#   # if( Max == p.bltr) symbol$extra <- c(2, "bltr")
#   # if( Max == p.tlbr) symbol$extra <- c(2, "tlbr")
#   # if( Max == p.tl) symbol$extra <- c(3, "tl")
#   # if( Max == p.tr) symbol$extra <- c(3, "tr")
#   # if( Max == p.br) symbol$extra <- c(3, "br")
#   # if( Max == p.bl) symbol$extra <- c(3, "bl")
#   # if( Max == p.four) 
#   symbol$extra <- c(4, "four")
# 
#   return( dens.biv.int(symbol, param=param, log=log) )
# 
# }



# Symbolic log-likelihood function

llik.biv.int.full <- function(symbol.data, param=c(0,0,1,0,1), log=TRUE, vectorial=FALSE){
  # symbol.data is the symbolic dataset, a list of symbols as defined in dens.biv.int
  # If vectorial=TRUE then a vector of each log-likelihood evaluation is returned, otherwise the sum is returned
  
  #cat("params: ", param, " \n")
  
  if( param[3]<=0 || param[5]<=0 ){
    if(log){
      return( -1e300 )
    }else{
      return( 1e-300)
    }
  }
  
  # Simulate the probabilities for the interval to be constructed from 2, 3, 4 points:
  
  n <- symbol.data[[1]]$n
  Mean <- param[1:2]
  Sigma <- matrix(c(param[3], param[4], param[4], param[5]), ncol=2)
  
  if(any(eigen(Sigma)$value<0)){ # Make sure the covariance matrix is positive definite
    if(log){
      return( -1e300 )
    }else{
      return( 1e-300)
    }
  }
    
  repeats <- 10000  
  n.bltr <- n.tlbr <- n.tl <- n.tr <- n.br <- n.bl <- n.four <- 0
  
  set.seed(14342)

  for(i in 1:repeats){

    tmp <- rmvnorm(n, mean = Mean, sigma = Sigma)
    if( which.max(tmp[,1]) == which.max(tmp[,2]) && which.min(tmp[,1]) == which.min(tmp[,2]) ) n.bltr <- n.bltr + 1
    if( which.max(tmp[,1]) == which.min(tmp[,2]) && which.min(tmp[,1]) == which.max(tmp[,2]) ) n.tlbr <- n.tlbr + 1
    if( which.min(tmp[,1]) == which.max(tmp[,2]) && which.max(tmp[,1]) != which.min(tmp[,2]) ) n.tl <- n.tl + 1
    if( which.max(tmp[,1]) == which.max(tmp[,2]) && which.min(tmp[,1]) != which.min(tmp[,2]) ) n.tr <- n.tr + 1
    if( which.max(tmp[,1]) == which.min(tmp[,2]) && which.min(tmp[,1]) != which.max(tmp[,2]) ) n.br <- n.br + 1
    if( which.min(tmp[,1]) == which.min(tmp[,2]) && which.max(tmp[,1]) != which.max(tmp[,2]) ) n.bl <- n.bl + 1
    if( which.max(tmp[,1]) != which.max(tmp[,2]) && which.min(tmp[,1]) != which.min(tmp[,2]) && which.max(tmp[,1]) != which.min(tmp[,2]) && which.min(tmp[,1]) != which.max(tmp[,2]) ) n.four <- n.four + 1
  }
  if( sum(n.bltr, n.tlbr, n.tl, n.tr, n.br, n.bl, n.four) != repeats){stop("Error in calculating the probabilities")}
  p.bltr <- n.bltr/repeats; p.tlbr <- n.tlbr/repeats;
  p.tl <- n.tl/repeats; p.tr <- n.tr/repeats; p.br <- n.br/repeats; p.bl <- n.bl/repeats;
  p.four <- n.four/repeats
  
  m <- length(symbol.data) # number of symbols
  llik <- vector("numeric",length=m)
  
  for( j in 1:m){
    llik[j] <- dens.biv.int.full(symbol=symbol.data[[j]], param=param, p.bltr=p.bltr, p.tlbr=p.tlbr, p.tl=p.tl, p.tr=p.tr, p.br=p.br, p.bl=p.bl, p.four=p.four, log=log)
  }
  
  if(vectorial){
    return(llik)
  }else{
    return(sum(llik))
  }
  
}  

# Optimisation of the log-likelihood function

fit.biv.int.full <- function(symbol.data, parastart, hessian=TRUE, trace=0){
  optimfun <- function(para){
    return(llik.biv.int.full(symbol.data=symbol.data, param=para, log=TRUE))
  }
  est.para <- optim(parastart, optimfun, method="BFGS", control = list(maxit = 2000, trace = trace, fnscale=-1), hessian=hessian)
  if(hessian){
    return(list(est=est.para$par, stderr=sqrt(diag(solve(-est.para$hessian))), hess=est.para$hessian, LL=est.para$value))
  }else{
    return(list(est=est.para$par ))
  }
  
}


fit.biv.int.fVar <- function(symbol.data, var, parastart, trace=0){
  optimfun <- function(para){
    return(llik.biv.int(symbol.data=symbol.data, param=c(para[1:2],var[1],para[3],var[2]), log=TRUE))
  }
  est.para <- optim(parastart, optimfun, method="BFGS", control = list(maxit = 2000, trace = trace, fnscale=-1), hessian=TRUE)
  return(list(est=est.para$par, stderr=sqrt(diag(solve(-est.para$hessian))), hess=est.para$hessian, LL=est.para$value))
}

fit.biv.int.fCov <- function(symbol.data, cov, parastart, trace=0){
  optimfun <- function(para){
    return(llik.biv.int(symbol.data=symbol.data, param=c(para[1:3],cov,para[4]), log=TRUE))
  }
  est.para <- optim(parastart, optimfun, method="BFGS", control = list(maxit = 2000, trace = trace, fnscale=-1), hessian=TRUE)
  return(list(est=est.para$par, stderr=sqrt(diag(solve(-est.para$hessian))), hess=est.para$hessian, LL=est.para$value))
}

fit.biv.int.Xmarg <- function(symbol.data, mean.y, cov, var.y, parastart, trace=0){
  optimfun <- function(para){
    return(llik.biv.int(symbol.data=symbol.data, param=c(para[1], mean.y, para[2],cov,var.y), log=TRUE))
  }
  est.para <- optim(parastart, optimfun, method="BFGS", control = list(maxit = 2000, trace = trace, fnscale=-1), hessian=TRUE)
  return(list(est=est.para$par, stderr=sqrt(diag(solve(-est.para$hessian))), hess=est.para$hessian, LL=est.para$value))
}

### For classical MLE

fit.mle <- function(class.data, parastart, trace=0){
  optimfun <- function(para){
    return(sum(dmvnorm(class.data, mean=para[1:2], sigma=matrix(c(para[3:4],para[4:5]), ncol=2), log=TRUE ) ))
  }
  est.para <- optim(parastart, optimfun, method="BFGS", control = list(maxit = 2000, trace = trace, fnscale=-1), hessian=TRUE)
  return(list(est=est.para$par, stderr=sqrt(diag(solve(-est.para$hessian))), hess=est.para$hessian, LL=est.para$value))
}


