### Likelihood function for bivariate intervals
###
###
### Intervals are given by four points:
###   (a1,b1), (a1,b2), (a2,b1) and (a2,b2)
### where a1 < a2 and b1 < b2
###
### Extra information is required: 
###   number of points used to construct the bivariate interval
###   can be 2, 3 or 4
###     If 2 points then we need to specify which ones:
###       "bltr" for bottom-left and top-right
###       "tlbr" for top-left and top-right
###     If 3 points we need to specify which point is in a corner
###       then the other two are on the opposite edges
###     If 4 points no extras required
###
### We focus on the bivariate Normal distribution

library(mvtnorm)



dens.biv.int <- function(symbol, param=c(0,0,1,0,1), log=TRUE){
  
  # symbol should be a list containing:
  # 1) a vector "coord" of length 4 with a1, a2, b1, b2
  # 2) a list "extra" containing an integer 2, 3 or 4 and corresponding information 
  # 3) an integer giving the number of points in the interval
  #
  # parameter vector should be of length 5: (mean1, mean2, var1, cov, var2)
  #   var1 and var2 should be positive
  
  if(!is.list(symbol)){stop("Symbol should be a list.")}
  if( length(symbol$coord) !=4){stop(" 'symbol$coord' should be of lentgh 4.")}
  if( all(symbol$extra[1] != 2:4) ){stop(" 'symbol$extra' should contain value 2, 3 or 4 as first element.")}
  if( symbol$extra[1] == 2 && all(symbol$extra[2] != c("bltr", "tlbr")) ){stop(" Second element of 'symbol$extra' incorrectly defined. ")}
  if( symbol$extra[1] == 3 && all(symbol$extra[2] != c("bl", "br", "tl", "tr")) ){stop(" Second element of 'symbol$extra' incorrectly defined. ")}
  if(length(param) != 5){stop("Wrong length of parameter vector")}
  if( param[3]<=0 || param[5]<=0 ){
    if(log){
      return( -1e300 )
    }else{
      return( 1e-300)
    }
  }
  
  n <- symbol$n
  Mean <- param[1:2]
  Sigma <- matrix(c(param[3], param[4], param[4], param[5]), ncol=2)
  a1 <- symbol$coord[1]; a2 <- symbol$coord[2]; 
  b1 <- symbol$coord[3]; b2 <- symbol$coord[4];
  
  if(any(eigen(Sigma)$value<0)){ # Make sure the covariance matrix is positive definite
    if(log){
      return( -1e300 )
    }else{
      return( 1e-300)
    }
  }
  
  # Compute the probability to be in the rectangle
  #p <- pmvnorm(lower=rep(-Inf,2), upper=c(a2,b2), mean=Mean, sigma=Sigma) - pmvnorm(lower=rep(-Inf,2), upper=c(a2,b1), mean=Mean, sigma=Sigma) - pmvnorm(lower=rep(-Inf,2), upper=c(a1,b2), mean=Mean, sigma=Sigma) + pmvnorm(lower=rep(-Inf,2), upper=c(a1,b1), mean=Mean, sigma=Sigma)
  p <- pmvnorm(lower=c(a1,b1), upper=c(a2,b2), mean=Mean, sigma=Sigma)
  
    
  if(symbol$extra[1] == 2){
    n.in <- n-2 # number of obs inside the interval
    if(symbol$extra[2] == "bltr"){
      d <- prod( dmvnorm( rbind(c(a1,b1),c(a2,b2))  , mean=Mean, sigma=Sigma) )
    }
    if(symbol$extra[2] == "tlbr"){
      d <- prod( dmvnorm( rbind(c(a1,b2),c(a2,b1))  , mean=Mean, sigma=Sigma) )
    }
  }
  if(symbol$extra[1] == 3){
    n.in <- n-3 # number of obs inside the interval
    Sd.cond.b <- sqrt( (prod(diag(Sigma)) - Sigma[1,2]^2) / Sigma[2,2] ) # conditional std. dev. x given y-axis (b coord)
    Sd.cond.a <- sqrt( (prod(diag(Sigma)) - Sigma[1,2]^2) / Sigma[1,1] ) # conditional std. dev. y given x-axis (a coord)
    if(symbol$extra[2] == "bl"){
      d1 <- dmvnorm( c(a1,b1) , mean=Mean, sigma=Sigma) # one point in bottom left of rectangle
      Mean.cond.b <- Mean[1] + Sigma[1,2]/Sigma[2,2]*(b2-Mean[2])
      Mean.cond.a <- Mean[2] + Sigma[1,2]/Sigma[1,1]*(a2-Mean[1])
      evalA <- a2; evalB <- b2
    }
    if(symbol$extra[2] == "tl"){
      d1 <- dmvnorm( c(a1,b2) , mean=Mean, sigma=Sigma) # one point in top left of rectangle
      Mean.cond.b <- Mean[1] + Sigma[1,2]/Sigma[2,2]*(b1-Mean[2])
      Mean.cond.a <- Mean[2] + Sigma[1,2]/Sigma[1,1]*(a2-Mean[1])
      evalA <- a2; evalB <- b1      
    }
    if(symbol$extra[2] == "br"){
      d1 <- dmvnorm( c(a2,b1) , mean=Mean, sigma=Sigma) # one point in bottom right of rectangle
      Mean.cond.b <- Mean[1] + Sigma[1,2]/Sigma[2,2]*(b2-Mean[2])
      Mean.cond.a <- Mean[2] + Sigma[1,2]/Sigma[1,1]*(a1-Mean[1])
      evalA <- a1; evalB <- b2      
    }
    if(symbol$extra[2] == "tr"){
      d1 <- dmvnorm( c(a2,b2) , mean=Mean, sigma=Sigma) # one point in top right of rectangle
      Mean.cond.b <- Mean[1] + Sigma[1,2]/Sigma[2,2]*(b1-Mean[2])
      Mean.cond.a <- Mean[2] + Sigma[1,2]/Sigma[1,1]*(a1-Mean[1])
      evalA <- a1; evalB <- b1      
    }    
    d2 <- (pnorm( a2, mean=Mean.cond.b, sd=Sd.cond.b ) - pnorm( a1, mean=Mean.cond.b, sd=Sd.cond.b)) * dnorm(evalB, mean=Mean[2], sd=sqrt(Sigma[2,2])) # one point on edge between (a1,-) and (a2,-)
    d3 <- (pnorm( b2, mean=Mean.cond.a, sd=Sd.cond.a ) - pnorm( b1, mean=Mean.cond.a, sd=Sd.cond.a)) * dnorm(evalA, mean=Mean[1], sd=sqrt(Sigma[1,1])) # one point on edge between (-,b1) and (-,b2)
    d <- d1*d2*d3
  }
  if(symbol$extra[1] == 4){
    n.in <- n-4 # number of obs inside the interval
    Mean.cond.a1 <- Mean[2] + Sigma[1,2]/Sigma[1,1]*(a1-Mean[1]) # conditional mean y given x = a1
    Mean.cond.a2 <- Mean[2] + Sigma[1,2]/Sigma[1,1]*(a2-Mean[1]) # conditional mean y given x = a2
    Mean.cond.b1 <- Mean[1] + Sigma[1,2]/Sigma[2,2]*(b1-Mean[2]) # conditional mean x given y = b1
    Mean.cond.b2 <- Mean[1] + Sigma[1,2]/Sigma[2,2]*(b2-Mean[2]) # conditional mean x given y = b2
    Sd.cond.b <- sqrt( (prod(diag(Sigma)) - Sigma[1,2]^2) / Sigma[2,2] ) # conditional std. dev. x given y-axis (b coord)
    Sd.cond.a <- sqrt( (prod(diag(Sigma)) - Sigma[1,2]^2) / Sigma[1,1] ) # conditional std. dev. y given x-axis (a coord)
    d1 <- (pnorm( a2, mean=Mean.cond.b1, sd=Sd.cond.b ) - pnorm( a1, mean=Mean.cond.b1, sd=Sd.cond.b )) * dnorm(b1, mean=Mean[2], sd=sqrt(Sigma[2,2])) # one point between a1 and a2 given y=b1
    d2 <- (pnorm( a2, mean=Mean.cond.b2, sd=Sd.cond.b ) - pnorm( a1, mean=Mean.cond.b2, sd=Sd.cond.b )) * dnorm(b2, mean=Mean[2], sd=sqrt(Sigma[2,2])) # one point between a1 and a2 given y=b2
    d3 <- (pnorm( b2, mean=Mean.cond.a1, sd=Sd.cond.a ) - pnorm( b1, mean=Mean.cond.a1, sd=Sd.cond.a )) * dnorm(a1, mean=Mean[1], sd=sqrt(Sigma[1,1])) # one point between b1 and b2 given x=a1
    d4 <- (pnorm( b2, mean=Mean.cond.a2, sd=Sd.cond.a ) - pnorm( b1, mean=Mean.cond.a2, sd=Sd.cond.a )) * dnorm(a2, mean=Mean[1], sd=sqrt(Sigma[1,1])) # one point between b1 and b2 given x=a2
    d <- d1*d2*d3*d4
  }  
  
  if(d==0) d <- 1e-250
   
  if(log){
    return( log(d) + n.in * log(p[1]) )
  }else{
    return( d * p[1]^(n.in) )
  }

}

# Symbolic log-likelihood function

llik.biv.int <- function(symbol.data, param=c(0,0,1,0,1), log=TRUE, vectorial=FALSE){
  # symbol.data is the symbolic dataset, a list of symbols as defined in dens.biv.int
  # If vectorial=TRUE then a vector of each log-likelihood evaluation is returned, otherwise the sum is returned
  
  
  m <- length(symbol.data) # number of symbols
  llik <- vector("numeric",length=m)
  
  for( j in 1:m){
    llik[j] <- dens.biv.int(symbol=symbol.data[[j]], param=param, log=log)
  }
  
  if(vectorial){
    return(llik)
  }else{
    return(sum(llik))
  }
  
}  

# Optimisation of the log-likelihood function

fit.biv.int <- function(symbol.data, parastart, hessian=TRUE, trace=0){
  optimfun <- function(para){
    return(llik.biv.int(symbol.data=symbol.data, param=para, log=TRUE))
  }
  est.para <- optim(parastart, optimfun, method="BFGS", control = list(maxit = 2000, trace = trace, fnscale=-1), hessian=hessian)
  if(hessian){
    return(list(est=est.para$par, stderr=sqrt(diag(solve(-est.para$hessian))), hess=est.para$hessian, LL=est.para$value))
  }else{
    return(list(est=est.para$par ))
  }
  
}


fit.biv.int.fVar <- function(symbol.data, var, parastart, trace=0){
  optimfun <- function(para){
    return(llik.biv.int(symbol.data=symbol.data, param=c(para[1:2],var[1],para[3],var[2]), log=TRUE))
  }
  est.para <- optim(parastart, optimfun, method="BFGS", control = list(maxit = 2000, trace = trace, fnscale=-1), hessian=TRUE)
  return(list(est=est.para$par, stderr=sqrt(diag(solve(-est.para$hessian))), hess=est.para$hessian, LL=est.para$value))
}

fit.biv.int.fCov <- function(symbol.data, cov, parastart, trace=0){
  optimfun <- function(para){
    return(llik.biv.int(symbol.data=symbol.data, param=c(para[1:3],cov,para[4]), log=TRUE))
  }
  est.para <- optim(parastart, optimfun, method="BFGS", control = list(maxit = 2000, trace = trace, fnscale=-1), hessian=TRUE)
  return(list(est=est.para$par, stderr=sqrt(diag(solve(-est.para$hessian))), hess=est.para$hessian, LL=est.para$value))
}

fit.biv.int.Xmarg <- function(symbol.data, mean.y, cov, var.y, parastart, trace=0){
  optimfun <- function(para){
    return(llik.biv.int(symbol.data=symbol.data, param=c(para[1], mean.y, para[2],cov,var.y), log=TRUE))
  }
  est.para <- optim(parastart, optimfun, method="BFGS", control = list(maxit = 2000, trace = trace, fnscale=-1), hessian=TRUE)
  return(list(est=est.para$par, stderr=sqrt(diag(solve(-est.para$hessian))), hess=est.para$hessian, LL=est.para$value))
}

### For classical MLE

fit.mle <- function(class.data, parastart, trace=0){
  optimfun <- function(para){
    return(sum(dmvnorm(class.data, mean=para[1:2], sigma=matrix(c(para[3:4],para[4:5]), ncol=2), log=TRUE ) ))
  }
  est.para <- optim(parastart, optimfun, method="BFGS", control = list(maxit = 2000, trace = trace, fnscale=-1), hessian=TRUE)
  return(list(est=est.para$par, stderr=sqrt(diag(solve(-est.para$hessian))), hess=est.para$hessian, LL=est.para$value))
}


