####
#### This file contains code to compute the density, likelihood and likliehood maximisation
####    by considering first the x-axis and y-axis
####    according to the sequential nesting method
#### 

###
### 
### Construction from the x-axis
###
###

dens.biv.os.int.x <- function(symbol, param=c(0,0,1,0,1), log=TRUE){
  
  # symbol should be a list containing:
  # 1) a vector "coord" of length 4 with x(l1), x(u1), y(l2), y(u2)
  # 2) a vector "orders" of length 4 with l1, u1, l2, u2
  # 3) a integer "n" giving the number of points aggregated
  #
  # parameter vector should be of length 5: (mean1, mean2, var1, cov, var2)
  #   var1 and var2 should be positive
  
  if(!is.list(symbol)){stop("Symbol should be a list.")}
  if( length(symbol$coord) !=4){stop(" 'symbol$coord' should be of lentgh 4.")}
  if(length(param) != 5){stop("Wrong length of parameter vector")}
  #if(missing(symbol$n)){stop(" missing 'n' in symbol ")}
  
  if( param[3]<=0 || param[5]<=0 ){
    if(log){
      return( -1e300 )
    }else{
      return( 1e-300)
    }
  }
  
  n <- symbol$n
  Mean <- param[1:2]
  Sigma <- matrix(c(param[3], param[4], param[4], param[5]), ncol=2)
  xl1 <- symbol$coord[1]; xu1 <- symbol$coord[2]; 
  yl2 <- symbol$coord[3]; yu2 <- symbol$coord[4];
  l1 <- symbol$orders[1]; u1 <- symbol$orders[2];
  l2 <- symbol$orders[3]; u2 <- symbol$orders[4];
  
  if(any(eigen(Sigma)$value<0)){ # Make sure the covariance matrix is positive definite
    if(log){
      return( -1e300 )
    }else{
      return( 1e-300)
    }
  }
  
  # Compute the probability to be in the rectangle
  p <- pmvnorm(lower=c(xl1,yl2), upper=c(xu1,yu2), mean=Mean, sigma=Sigma)[1]

  # Compute the probablity of X to be less than x(l1)
  p.less.l1 <- pnorm(xl1, mean=Mean[1], sd=sqrt(Sigma[1,1]))

  # Compute the probablity of X to be greater than x(u1)
  p.more.u1 <- 1 - pnorm(xu1, mean=Mean[1], sd=sqrt(Sigma[1,1]))

  # Compute the probability of X to be between x(l1) and x(u1) and Y less than y(l2)  
  p.less.l2 <- pmvnorm(lower=c(xl1,-Inf), upper=c(xu1,yl2), mean=Mean, sigma=Sigma)[1]

  # Compute the probability of X to be between x(l1) and x(u1) and Y greater than y(u2)
  p.more.u2 <- pmvnorm(lower=c(xl1,yu2), upper=c(xu1,Inf), mean=Mean, sigma=Sigma)[1]

  # Contribution of the two points on the vertical axis x=x(l1) and x=x(u1)
  p.x <- prod( dnorm(c(xl1, xu1), mean=Mean[1], sd=sqrt(Sigma[1,1])) )

  # Contribution of the point on horizontal axis between x(l1) and x(u1) given y(l2)
  Cond.mean.l2 <- Mean[1] + Sigma[1,2]/Sigma[2,2]*(yl2-Mean[2])
  Sd.cond.y <- sqrt( (prod(diag(Sigma)) - Sigma[1,2]^2) / Sigma[2,2] )
  p.yl2 <- ( pnorm(xu1, mean=Cond.mean.l2, sd=Sd.cond.y) - pnorm(xl1, mean=Cond.mean.l2, sd=Sd.cond.y)  ) * dnorm(yl2, mean=Mean[2], sd=sqrt(Sigma[2,2]))

  # Contribution of the point on horizontal axis between x(l1) and x(u1) given y(u2)
  Cond.mean.u2 <- Mean[1] + Sigma[1,2]/Sigma[2,2]*(yu2-Mean[2])
  p.yu2 <- ( pnorm(xu1, mean=Cond.mean.u2, sd=Sd.cond.y) - pnorm(xl1, mean=Cond.mean.u2, sd=Sd.cond.y)  ) * dnorm(yu2, mean=Mean[2], sd=sqrt(Sigma[2,2]))

  # if( p<=0 || p.less.l1<=0 || p.more.u1<=0 || p.less.l2<=0 || p.more.u2<=0 || p.x<=0 || p.yl2<=0 || p.yu2<=0){
  #   if(log){
  #     return( -1e300 )
  #   }else{
  #     return( 1e-300)  
  #   }  
  # }
  
  if(log){
    return( (l1-1) * log(p.less.l1) +  (n-u1) * log(p.more.u1) + (l2-1) * log(p.less.l2) + (u1-u2-l1-1) * log(p.more.u2) + (u2-l2-1) * log(p) +  log(p.x) + log(p.yl2) + log(p.yu2) )
  }else{
    return( p.less.l1^(l1-1) * p.more.u1^(n-u1) * p.less.l2^(l2-1) * p.more.u2^(u1-u2-l1-1) * p^(u2-l2-1) * p.x * p.yl2 * p.yu2 )
  }
  
}

# Symbolic log-likelihood function

llik.biv.os.int.x <- function(symbol.data, param=c(0,0,1,0,1), log=TRUE, vectorial=FALSE){
  # symbol.data is the symbolic dataset, a list of symbols as defined in dens.biv.int
  # If vectorial=TRUE then a vector of each log-likelihood evaluation is returned, otherwise the sum is returned
  
  m <- length(symbol.data) # number of symbols
  llik <- vector("numeric",length=m)
  
  for( j in 1:m){
    llik[j] <- dens.biv.os.int.x(symbol=symbol.data[[j]], param=param, log=log)
  }
  
  if(vectorial){
    return(llik)
  }else{
    return(sum(llik))
  }
  
}  

# Optimisation of the log-likelihood function

fit.biv.os.int.x <- function(symbol.data, parastart, hessian=TRUE, trace=0){
  optimfun <- function(para){
    #cat("param:", para, "\n")
    return(llik.biv.os.int.x(symbol.data=symbol.data, param=para, log=TRUE))
  }
  est.para <- optim(parastart, optimfun, method="BFGS", control = list(maxit = 2000, trace = trace, fnscale=-1), hessian=hessian)
  if(hessian){
    return(list(est=est.para$par, stderr=sqrt(diag(solve(-est.para$hessian))), hess=est.para$hessian, LL=est.para$value))
  }else{
    return(list(est=est.para$par ))
  }
  
}

###
###
### Construction from the y-axis
###
###

dens.biv.os.int.y <- function(symbol, param=c(0,0,1,0,1), log=TRUE){
  
  # symbol should be a list containing:
  # 1) a vector "coord" of length 4 with x(l1), x(u1), y(l2), y(u2)
  # 2) a vector "orders" of length 4 with l1, u1, l2, u2
  # 3) a integer "n" giving the number of points aggregated
  #
  # parameter vector should be of length 5: (mean1, mean2, var1, cov, var2)
  #   var1 and var2 should be positive
  
  if(!is.list(symbol)){stop("Symbol should be a list.")}
  if( length(symbol$coord) !=4){stop(" 'symbol$coord' should be of lentgh 4.")}
  if(length(param) != 5){stop("Wrong length of parameter vector")}
  #if(missing(symbol$n)){stop(" missing 'n' in symbol ")}
  
  if( param[3]<=0 || param[5]<=0 ){
    if(log){
      return( -1e300 )
    }else{
      return( 1e-300)
    }
  }
  
  n <- symbol$n
  Mean <- param[1:2]
  Sigma <- matrix(c(param[3], param[4], param[4], param[5]), ncol=2)
  xl1 <- symbol$coord[1]; xu1 <- symbol$coord[2]; 
  yl2 <- symbol$coord[3]; yu2 <- symbol$coord[4];
  l1 <- symbol$orders[1]; u1 <- symbol$orders[2];
  l2 <- symbol$orders[3]; u2 <- symbol$orders[4];
  
  if(any(eigen(Sigma)$value<0)){ # Make sure the covariance matrix is positive definite
    if(log){
      return( -1e300 )
    }else{
      return( 1e-300)
    }
  }
  
  # Compute the probability to be in the rectangle
  p <- pmvnorm(lower=c(xl1,yl2), upper=c(xu1,yu2), mean=Mean, sigma=Sigma)[1]
  
  # Compute the probablity of X to be less than y(l2)
  p.less.l2 <- pnorm(yl2, mean=Mean[2], sd=sqrt(Sigma[2,2]))
  
  # Compute the probablity of X to be greater than x(u1)
  p.more.u2 <- 1 - pnorm(yu2, mean=Mean[2], sd=sqrt(Sigma[2,2]))
  
  # Compute the probability of X to be between y(l2) and y(u2) and X less than x(l1)  
  p.less.l1 <- pmvnorm(lower=c(-Inf,yl2), upper=c(xl1,yu2), mean=Mean, sigma=Sigma)[1]
  
  # Compute the probability of X to be between y(l2) and y(u2) and X greater than x(u1)
  p.more.u1 <- pmvnorm(lower=c(xu1,yl2), upper=c(Inf,yu2), mean=Mean, sigma=Sigma)[1]
  
  # Contribution of the two points on the horizontal axis y=y(l2) and y=y(u2)
  p.y <- prod( dnorm(c(yl2, yu2), mean=Mean[2], sd=sqrt(Sigma[2,2])) )
  
  # Contribution of the point on horizontal axis between y(l2) and y(u2) given x(l1)
  Cond.mean.l1 <- Mean[2] + Sigma[1,2]/Sigma[1,1]*(xl1-Mean[1])
  Sd.cond.x <- sqrt( (prod(diag(Sigma)) - Sigma[1,2]^2) / Sigma[1,1] )
  p.xl1 <- ( pnorm(yu2, mean=Cond.mean.l1, sd=Sd.cond.x) - pnorm(yl2, mean=Cond.mean.l1, sd=Sd.cond.x)  ) * dnorm(xl1, mean=Mean[1], sd=sqrt(Sigma[1,1]))
  
  # Contribution of the point on horizontal axis between y(l2) and y(u2) given x(u1)
  Cond.mean.u1 <- Mean[2] + Sigma[1,2]/Sigma[1,1]*(xu1-Mean[1])
  p.xu1 <- ( pnorm(yu2, mean=Cond.mean.u1, sd=Sd.cond.x) - pnorm(yl2, mean=Cond.mean.u1, sd=Sd.cond.x)  ) * dnorm(xu1, mean=Mean[1], sd=sqrt(Sigma[1,1]))
  
  # if( p<=0 || p.less.l1<=0 || p.more.u1<=0 || p.less.l2<=0 || p.more.u2<=0 || p.x<=0 || p.yl2<=0 || p.yu2<=0){
  #   if(log){
  #     return( -1e300 )
  #   }else{
  #     return( 1e-300)  
  #   }  
  # }
  
  if(log){
    return( (l2-1) * log(p.less.l2) +  (n-u2) * log(p.more.u2) + (l1-1) * log(p.less.l1) + (u2-u1-l2-1) * log(p.more.u1) + (u1-l1-1) * log(p) +  log(p.y) + log(p.xl1) + log(p.xu1) )
  }else{
    return( p.less.l2^(l2-1) * p.more.u2^(n-u2) * p.less.l1^(l1-1) * p.more.u1^(u2-u1-l2-1) * p^(u1-l1-1) * p.y * p.xl1 * p.xu1 )
  }
  
}

# Symbolic log-likelihood function

llik.biv.os.int.y <- function(symbol.data, param=c(0,0,1,0,1), log=TRUE, vectorial=FALSE){
  # symbol.data is the symbolic dataset, a list of symbols as defined in dens.biv.int
  # If vectorial=TRUE then a vector of each log-likelihood evaluation is returned, otherwise the sum is returned
  
  m <- length(symbol.data) # number of symbols
  llik <- vector("numeric",length=m)
  
  for( j in 1:m){
    llik[j] <- dens.biv.os.int.y(symbol=symbol.data[[j]], param=param, log=log)
  }
  
  if(vectorial){
    return(llik)
  }else{
    return(sum(llik))
  }
  
}  

# Optimisation of the log-likelihood function

fit.biv.os.int.y <- function(symbol.data, parastart, hessian=TRUE, trace=0){
  optimfun <- function(para){
    #cat("param:", para, "\n")
    return(llik.biv.os.int.y(symbol.data=symbol.data, param=para, log=TRUE))
  }
  est.para <- optim(parastart, optimfun, method="BFGS", control = list(maxit = 2000, trace = trace, fnscale=-1), hessian=hessian)
  if(hessian){
    return(list(est=est.para$par, stderr=sqrt(diag(solve(-est.para$hessian))), hess=est.para$hessian, LL=est.para$value))
  }else{
    return(list(est=est.para$par ))
  }
  
}

