###
###
### Functions for the hierarchical model
###		replicating the results from Le Rademacher and Billard (2011)
###		
###


dens.billard <- function(symbol, type, param, log){
  
  # Param is a vector of length 5:
  #   5 parameters for the mean which is t distributed (location is quadratic in risk, scale and df)
  #   2 parameters for the variance which is inverse-gamma distributed (shape and scale)
  
  if( all(type != c("hist", "int" )) ){stop("Wrong symbol type")}
  if(length(param) != 7){ stop("Wrong length of parameter vector")}
  
  a = param[1]; b = param[2]; c = param[3]; 
  ohm = param[4]; nu = param[5];  
  shape = param[6]; scale = param[7];
  
  if(ohm <=0 || nu <= 0 || shape <= 0 || scale <= 0){if(log){return(-1e300)}else{return(1e-300)}}
  
  if(type=="int"){ #INTERVAL
    
    #if(length(symbol$interv) != 2){stop("Wrong length of interval symbol")}
    
    mean = sum(symbol$interv[1:2])/2
    var = diff(symbol$interv[1:2])^2/12
    
  }
  
  if(type=="hist"){ #HISTOGRAM
    #if( !exists(symbol$breaks) || !exists(symbol$counts) ){ stop("Histogram symbols needs 'breaks' and 'counts'") }
    n <- sum(symbol$counts);
    # m <- length(symbol$counts);
    
    A <- head(symbol$breaks, -1)
    B <- tail(symbol$breaks, -1)
    
    mean = sum( (A+B)/2 * symbol$counts ) / n
    var  =  sum( (A^2 + A*B + B^2) * symbol$counts ) / (3*n) - mean^2
    
  }
  
  d1 <- dt(x=c(mean - a * symbol$risk.lvl^2 - b * symbol$risk.lvl - c)/sqrt(ohm), df=nu, log=log )
  if(log) d11 <- log(ohm)/2 else d11 <- sqrt(ohm)
  if(d1==-Inf) d1 <- -1e300
  d2 <- dinvgamma(x=var, shape=shape, scale=scale)
  
  if(log) return(d1 - d11 + log(d2)) else return(d1 / d11 * d2)
  
}


llik.billard <- function(symbol.data, type, param, log=TRUE, vectorial=FALSE){

  #cat("Params: ", param, "\n")
  m <- length(symbol.data) # number of symbols
  llik <- vector("numeric",length=m)
  
  for( j in 1:m){
    llik[j] <- dens.billard(symbol=symbol.data[[j]], type=type, param=param, log=log)
  }
  
  if(vectorial){
    return(llik)
  }else{
    if(log) return(sum(llik)) else return(prod(llik))
  }
  
}  


fit.billard <- function(symbol.data, type, parastart, method="Nelder-Mead", hessian=TRUE, trace=0,...){

  optimfun <- function(para){
    return(llik.billard(symbol.data=symbol.data, type=type, param=para))
  }
  est.para <- optim(parastart, optimfun, method=method, control = list(maxit = 2e+7, trace = trace, fnscale=-1,...), hessian=hessian)
  if(hessian){
    return(list(est=est.para$par, stderr=sqrt(diag(solve(-est.para$hessian))), hess=est.para$hessian, LL=est.para$value))
  }else{
    return(list(est=est.para$par, LL= est.para$value ))
  }
  
}

fit.billard.fDF <- function(symbol.data, type, parastart, df, method="Nelder-Mead", hessian=TRUE, trace=0,...){
  
  optimfun <- function(para){
    return(llik.billard(symbol.data=symbol.data, type=type, param=c(para[1:4], df, para[5:6])) )
  }
  est.para <- optim(parastart, optimfun, method=method, control = list(maxit = 2e+7, trace = trace, fnscale=-1,...), hessian=hessian)
  if(hessian){
    return(list(est=est.para$par, stderr=sqrt(diag(solve(-est.para$hessian))), hess=est.para$hessian, LL=est.para$value))
  }else{
    return(list(est=est.para$par, LL= est.para$value ))
  }
  
}


dens.billard2 <- function(symbol, param, log){
  
  # Param is a vector of length 5:
  #   5 parameters for the mean which is t distributed (location is quadratic in risk, scale and df)
  #   2 parameters for the variance which is inverse-gamma distributed (shape and scale)
  
  if(length(param) != 7){ stop("Wrong length of parameter vector")}
  
  mean= symbol$val[1]; var=symbol$val[2];
  
  a = param[1]; b = param[2]; c = param[3]; 
  ohm = param[4]; nu = param[5];  
  shape = param[6]; scale = param[7];
  
  if(ohm <=0 || nu <= 0 || shape <= 0 || scale <= 0){if(log){return(-1e300)}else{return(1e-300)}}
  
  d1 <- dt(x=c(mean - a * symbol$risk.lvl^2 - b * symbol$risk.lvl - c)/sqrt(ohm), df=nu, log=log )
  if(log) d11 <- log(ohm)/2 else d11 <- sqrt(ohm)
  if(d1==-Inf) d1 <- -1e300
  d2 <- dinvgamma(x=var, shape=shape, scale=scale)
  
  if(log) return(d1 - d11 + log(d2)) else return(d1 / d11 * d2)
  
}


llik.billard2 <- function(symbol.data, param, log=TRUE, vectorial=FALSE){
  
  #cat("Params: ", param, "\n")
  m <- length(symbol.data) # number of symbols
  llik <- vector("numeric",length=m)
  
  for( j in 1:m){
    llik[j] <- dens.billard2(symbol=symbol.data[[j]], param=param, log=log)
  }
  
  if(vectorial){
    return(llik)
  }else{
    if(log) return(sum(llik)) else return(prod(llik))
  }
  
}  


