###
###
### Functions for the classical hierarchical model
###		assuming Normal and Skew-Normal distributions
###

###################################################################
########## NORMAL DISTRIBUTION ####################################
###################################################################

########## Inclusion of the distribution of the scale #############

dens.hier.univ.normS <- function(data, risk.lvl, theta, varphi, log){
  
  # "data" is the classical univariate observations in a specific risk.lvl group
  # "risk.lvl" constaining a value from 1 to 35 correspond to the risk level A1 to G5
  # "theta" a vector of parameters: location and scale, length is 2 
  # "varphi" is a vector of parameters for the distribution of the mean and variance
  #        - The location (univariate) is T distributed with:
  #             -- mean as function of the risk level, "a risk^2 + b risk + c" and "d risk^2 + e risk + f";
  #             -- variance, "tau" a real times the scale matrix;
  #             -- degree of freedom "nu";
  #        - The variance is Inverse-gamma distributed  with:
  #             -- shape;
  #             -- rate;
  # "log" to decide if the density or the log-density is returned
  
  if(length(varphi) != 7){ stop("Wrong length of parameter varphi") }
  if(!is.integer(risk.lvl) || risk.lvl < 1 || risk.lvl > 35){ stop("risk.lvl must be an integer between 1 and 35") }
  
  a = varphi[1]; b = varphi[2]; c = varphi[3]; 
  ohm = varphi[4]; nu = varphi[5];  
  shape = varphi[6]; scale = varphi[7];
  
  if(ohm <=0 || nu <= 0 || shape <= 0 || scale <= 0){if(log){return(-1e300)}else{return(1e-300)}}
  
  x.bar <- data[1] # mean data
  s <- data[2] # sd data
  n <- data[3] #length data
  d11 <- dnorm(x.bar, mean = theta[1], sd = sqrt(theta[2])/sqrt(n), log = log)
  d12 <- dnorm(s, sd = sqrt(theta[2])/sqrt(n-1), log = log)
  if(log) d13 <- (-n/2 + 1) * log(2*pi*theta[2]) - 0.5 * log(n*(n-1)) else d13 <- (2*pi*theta[2])^(-n/2 + 1) / sqrt(n*(n-1))
  if(log) d1 <- d11 + d12 + d13 else d1 <- d11 * d12 * d13
  
  d2 <- dt(x=c(theta[1]- a * risk.lvl^2 - b * risk.lvl - c)/sqrt(ohm), df=nu, log=log )
  if(log) d3 <- log(ohm)/2 else d3 <- sqrt(ohm)
  if(d2==-Inf) d2 <- -1e300
  
  d4 <- dinvgamma(x=theta[2], shape=shape, scale=scale)
  
  if(log) return(sum(d1) + d2 - d3 + log(d4)) else return(prod(d1) * d2 / d3 * d4)
  
}

llik.hier.univ.normS <- function(data, risk, Theta, varphi, log=TRUE, vectorial=FALSE){
  # If vectorial=TRUE then a vector of each log-likelihood evaluation is returned, otherwise the sum is returned
  #
  # Theta is a vector of all the theta_i, i being the group index
  # so for the Normal model each theta is of length 5 and thus Theta is of length m*5
  
  m <- length(data) # number of risk groups
  llik <- vector("numeric",length=m)
  
  if(length(Theta) != m*2){stop("Wrong lentgh of Theta vector")}
    
  for( j in 1:m){
    llik[j] <- dens.hier.univ.normS(data=data[[j]]$data, risk.lvl = data[[j]]$risk.lvl, theta=Theta[((data[[j]]$risk.lvl-1)*2+1):(data[[j]]$risk.lvl*2)], varphi=varphi, log=log)
  }
  
  if(vectorial){
    return(llik)
  }else{
    if(log) return(sum(llik)) else return(prod(llik))
  }
  
}  

fit.hier.univ.normS <- function(data, parastart, hessian=TRUE, method="Nelder-Mead", trace=0,...){
  m <- length(data) # number of risk levels
  optimfun <- function(para){
    return(llik.hier.univ.normS(data=data, Theta=para[1:(m*2)], varphi=para[(m*2 +1):(m*2 + 7)], log=TRUE))
  }
  est.para <- optim(parastart, optimfun, method=method, control = list(maxit = 2e+7, trace = trace, fnscale=-1,...), hessian=hessian)
  if(hessian){
    return(list(est=est.para$par, stderr=sqrt(diag(solve(-est.para$hessian))), hess=est.para$hessian, LL=est.para$value))
  }else{
    return(list(est=est.para$par, LL= est.para$value ))
  }
  
}

fit.hier.univ.normS.fDF <- function(data, parastart, df, hessian=TRUE, method="Nelder-Mead", trace=0,...){
  m <- length(data) # number of risk levels
  optimfun <- function(para){
    return(llik.hier.univ.normS(data=data, Theta=para[1:(m*2)], varphi=c(para[(m*2 +1):(m*2 + 4)], df, para[(m*2 +5):(m*2 + 6)] ), log=TRUE))
  }
  est.para <- optim(parastart, optimfun, method=method, control = list(maxit = 2e+7, trace = trace, fnscale=-1,...), hessian=hessian)
  if(hessian){
    return(list(est=est.para$par, stderr=sqrt(diag(solve(-est.para$hessian))), hess=est.para$hessian, LL=est.para$value))
  }else{
    return(list(est=est.para$par, LL= est.para$value ))
  }
  
}


###################################################################
########## SKEW-NORMAL DISTRIBUTION ###############################
###################################################################

########## Inclusion of the distribution of the scale #############
########## Inclusion of the distribution of the skewness ##########

dens.hier.univ.SKnormSS <- function(data, risk.lvl, theta, parametrisation="dp", varphi, log){
  
  # "data" is the classical univariate observations in a specific risk.lvl group
  # "risk.lvl" constaining a value from 1 to 35 correspond to the risk level A1 to G5
  # "theta" a vector of parameters: location, scale and skewness, length is 3 
  # "varphi" is a vector of parameters for the distribution of the mean and variance
  #        - The location (univariate) is T distributed with:
  #             -- mean as function of the risk level, "a risk^2 + b risk + c" and "d risk^2 + e risk + f";
  #             -- variance, "tau" a real times the scale matrix;
  #             -- degree of freedom "nu";
  #        - The variance is Inverse-gamma distributed  with:
  #             -- shape;
  #             -- rate;
  #        - The skewness is Normally distributed with:
  #             -- mean "mu.sk";
  #             -- variance "sig.sk";
  # "log" to decide if the density or the log-density is returned
  
  if(length(varphi) != 9){ stop("Wrong length of parameter varphi") }
  if(!is.integer(risk.lvl) || risk.lvl < 1 || risk.lvl > 35){ stop("risk.lvl must be an integer between 1 and 35") }
  
  if(length(theta) != 3){stop("Wrong length of parameters")}
  if(theta[2] <=0){if(log){return(-1e300)}else{return(1e-300)}}
  
  a = varphi[1]; b = varphi[2]; c = varphi[3]; 
  ohm = varphi[4]; nu = varphi[5];  
  shape = varphi[6]; scale = varphi[7];
  mu.sk = varphi[8]; sig.sk = varphi[9];
  
  if(ohm <=0 || nu <= 0 || shape <= 0 || scale <= 0 || sig.sk <= 0){if(log){return(-1e300)}else{return(1e-300)}}
  
  if(parametrisation=="dp") dp <-c(theta[1], sqrt(theta[2]), theta[3])
  if(parametrisation=="cp"){
    if(abs(theta[3]) >= (4-pi)/2 * (2 / (pi-2))^(3/2) ){if(log){return(-1e300)}else{return(1e-300)}}
    dp <- cp2dp(c(theta[1], sqrt(theta[2]), theta[3]) , "SN")
  }
  
  d1 <- dsn(x=data, dp=dp, log=log)

  d2 <- dt(x=c(theta[1]- a * risk.lvl^2 - b * risk.lvl - c)/sqrt(ohm), df=nu, log=log )
  if(log) d3 <- log(ohm)/2 else d3 <- sqrt(ohm)
  if(d2==-Inf) d2 <- -1e300
  
  d4 <- dinvgamma(x=theta[2], shape=shape, scale=scale)
  d5 <- dnorm(x=theta[3], mean=mu.sk, sd=sqrt(sig.sk), log=log)
  
  if(log) return(sum(d1) + d2 - d3 + log(d4) + d5) else return(prod(d1) * d2 / d3 * d4 * d5)
  
  # return(d1)
}

llik.hier.univ.SKnormSS <- function(data, risk, Theta, parametrisation="dp", varphi, log=TRUE, vectorial=FALSE, parallel=FALSE){
  # If vectorial=TRUE then a vector of each log-likelihood evaluation is returned, otherwise the sum is returned
  #
  # Theta is a vector of all the theta_i, i being the group index
  # so for the Normal model each theta is of length 5 and thus Theta is of length m*5
  
  #cat("Params: ", Theta, "\n")
  m <- length(data) # number of risk groups
  
  if(length(Theta) != m*3){stop("Wrong lentgh of Theta vector")}
  
  # llik <- apply(cbind(data,risk),1, function(x) dens.hier.norm(data=x[1:2], risk.lvl = as.integer(x[3]), theta=Theta[((risk.lvl-1)*5+1):(risk.lvl*5)], varphi=varphi, log=log) )
  
  if(parallel){
    llik <- foreach(i=1:m, .combine = c, .export = c("dens.hier.univ.SKnormSS")) %dopar% {
      dens.hier.univ.SKnormSS(data=data[[i]]$data, risk.lvl = data[[i]]$risk.lvl, theta=Theta[((data[[i]]$risk.lvl-1)*3+1):(data[[i]]$risk.lvl*3)], parametrisation=parametrisation, varphi=varphi, log=log)
    }
  }else{
    llik <- vector("numeric",length=m)
    for( j in 1:m){
      llik[j] <- dens.hier.univ.SKnormSS(data=data[[j]]$data, risk.lvl = data[[j]]$risk.lvl, theta=Theta[((data[[j]]$risk.lvl-1)*3+1):(data[[j]]$risk.lvl*3)], parametrisation=parametrisation, varphi=varphi, log=log)
    }
  }
  
  if(vectorial){
    return(llik)
  }else{
    if(log) return(sum(llik)) else return(prod(llik))
  }
  
}  

fit.hier.univ.SKnormSS <- function(data, parastart, parametrisation="dp", parallel=FALSE, hessian=TRUE, method="Nelder-Mead", trace=0,...){
  m <- length(data) # number of risk levels
  optimfun <- function(para){
    return(llik.hier.univ.SKnormSS(data=data, Theta=para[1:(m*3)], parametrisation=parametrisation, varphi=para[(m*3 +1):(m*3 + 9)], log=TRUE, parallel=parallel))
  }
  est.para <- optim(parastart, optimfun, method=method, control = list(maxit = 2e+7, trace = trace, fnscale=-1,...), hessian=hessian)
  if(hessian){
    return(list(est=est.para$par, stderr=sqrt(diag(solve(-est.para$hessian))), hess=est.para$hessian, LL=est.para$value))
  }else{
    return(list(est=est.para$par, LL= est.para$value ))
  }
  
}

fit.hier.univ.SKnormSS.fDF <- function(data, parastart, parametrisation="dp", df, parallel=FALSE, hessian=TRUE, method="Nelder-Mead", trace=0,...){
  m <- length(data) # number of risk levels
  optimfun <- function(para){
    return(llik.hier.univ.SKnormSS(data=data, Theta=para[1:(m*3)], parametrisation=parametrisation, varphi=c(para[(m*3 +1):(m*3 + 4)], df, para[(m*3 +5):(m*3 + 8)] ), log=TRUE, parallel=parallel))
  }
  est.para <- optim(parastart, optimfun, method=method, control = list(maxit = 2e+7, trace = trace, fnscale=-1,...), hessian=hessian)
  if(hessian){
    return(list(est=est.para$par, stderr=sqrt(diag(solve(-est.para$hessian))), hess=est.para$hessian, LL=est.para$value))
  }else{
    return(list(est=est.para$par, LL= est.para$value ))
  }
  
}

