###
###
### Functions for the hierarchical model
###		for univariate histogram symbols
###		assuming Normal and Skew-Normal distributions
###

### Multinomial coefficient is not including as it is not needed in the parameter estimation

###
###
###
### density univariate histogram for both normal and skew-normal distributions
###
###
###

dsymbolic1 <- function(counts=NULL, breaks=NULL, distribution=NULL, param, parametrisation="dp", log=FALSE){
  
  if( length(counts) != (length(breaks)-1) ){stop("Wrong breaks length") }
  if(is.null(counts) || is.null(breaks) || is.null(distribution)){stop("Missing input") }
  
  S=sum(counts)
  B <- length(counts)
  
  if(log){p <- 0}else{p <- 1}
  
  ## For underlying normal distribution
  
  if(distribution=='Normal'){
    
    if(length(param) != 2){stop("Wrong length of parameters")}
    if(param[2] <=0 ){if(log){return(log(-1e320))}else{return(1e-320)}}

    # Compute all the probabilities
    n <- length(breaks); 
    Probs <- pnorm(breaks, mean=param[1], sd=sqrt(param[2]))
    
    for(i in 1:B){
      prob <- Probs[i+1] - Probs[i]
      if(prob <= 0){prob <- 1e-320}
      if(log){
        p <- p + log( prob ) * counts[i]
      }else{
        p <- p * ( prob )^counts[i]
      }  
    }
      
  } # END: if distribution=="Normal"
  
  ## For underlying skew-normal distribution
  
  if(distribution=='Sk-Normal'){
    
    if(length(param) != 3){stop("Wrong length of parameters")}
    if(param[2] <=0){if(log){return(-1e300)}else{return(1e-300)}}

    if(parametrisation=="dp") dp <-c(param[1], sqrt(param[2]), param[3])
    if(parametrisation=="cp"){
      if(abs(param[3]) >= (4-pi)/2 * (2 / (pi-2))^(3/2) ){if(log){return(-1e300)}else{return(1e-300)}}
      dp <- cp2dp(c(param[1], sqrt(param[2]), param[3]) , "SN")
    } 
   
    # Compute all the probabilities
    n <- length(breaks); 
    Probs <- psn(breaks, dp=dp)
    
    for(i in 1:B){
      prob <- Probs[i+1] - Probs[i]
      if(prob <= 0){prob <- 1e-320}
      if(log){
        p <- p + log( prob ) * counts[i]
      }else{
        p <- p * ( prob )^counts[i]
      }  
    }

  }  
  
  return(p)	
}


###
###
###
### FUNCTIONS FOR HIERARCHICAL MODEL
###
###
###


###################################################################
########## NORMAL DISTRIBUTION ####################################
###################################################################

########## Inclusion of the distribution of the scale #############

dens.hier.univ.hist.normS <- function(symbol, theta, varphi, log){
  
  # "symbol" is a list containing elements:
  #        - 'counts' 
  #        - 'breaks'
  #        - 'risk.lvl' constaining a value from 1 to 35 correspond to the risk level A1 to G5
  # "theta" a vector of parameters: location and scale (var), length is 2 
  # "varphi" is a vector of parameters for the distribution of the mean and variance
  #        - The location is T distributed with:
  #             -- mean as function of the risk level, "a risk + b" and "c risk + d";
  #             -- variance;
  #             -- degree of freedom "nu";
  #        - The scale matrix is Inverse-Gamma distributed  with:
  #             -- degree of freedom "v";
  #             -- a scale matrix "S";
  # "log" to decide if the density or the log-density is returned
  
  if(length(varphi) != 7){ stop("Wrong length of parameter varphi") }
  if(!is.integer(symbol$risk.lvl) || symbol$risk.lvl < 1 || symbol$risk.lvl > 35){ stop("risk.lvl must be an integer between 1 and 35") }
  
  a = varphi[1]; b = varphi[2]; c = varphi[3]; 
  ohm = varphi[4]; nu = varphi[5];  
  shape = varphi[6]; scale = varphi[7];
  
  if(ohm <=0 || nu <= 0 || shape <= 0 || scale <= 0){if(log){return(-1e300)}else{return(1e-300)}}
  
  d1 <- dsymbolic1(counts=symbol$counts, breaks=symbol$breaks, distribution="Normal", param=theta, log=log)
  d2 <- dt(x=c(theta[1]- a * symbol$risk.lvl^2 - b * symbol$risk.lvl - c)/sqrt(ohm), df=nu, log=log )
  if(log) d3 <- log(ohm)/2 else d3 <- sqrt(ohm)
  
  if(d2==-Inf) d2 <- -1e300
  d4 <- dinvgamma(x=theta[2], shape=shape, scale=scale)
  
  if(log) return(d1 + d2 - d3 + log(d4)) else return(d1 * d2 / d3 * d4)
  
  # return(d1)
}


llik.hier.univ.hist.normS <- function(symbol.data, Theta, varphi, log=TRUE, vectorial=FALSE){
  # symbol.data is the symbolic dataset, a list of symbols as defined in dens.hier.biv.int.norm
  # If vectorial=TRUE then a vector of each log-likelihood evaluation is returned, otherwise the sum is returned
  #
  # Theta is a vector of all the theta_i, i being the group index
  # so for the Normal model each theta is of length 5 and thus Theta is of length m*5
  
  #cat("Params: ", Theta, "\n")
  m <- length(symbol.data) # number of symbols
  llik <- vector("numeric",length=m)
  
  if(length(Theta) != m*2){stop("Wrong lentgh of Theta vector")}
  
  for( j in 1:m){
    llik[j] <- dens.hier.univ.hist.normS(symbol=symbol.data[[j]], theta=Theta[((j-1)*2+1):(j*2)], varphi=varphi, log=log)
  }
  
  if(vectorial){
    return(llik)
  }else{
    if(log) return(sum(llik)) else return(prod(llik))
  }
  
}  


fit.hier.univ.hist.normS <- function(symbol.data, parastart, method="Nelder-Mead", hessian=TRUE, trace=0,...){
  m <- length(symbol.data) # number of symbols
  optimfun <- function(para){
    return(llik.hier.univ.hist.normS(symbol.data=symbol.data, Theta= para[1:(m*2)], varphi=para[(m*2 +1):(m*2 + 7)], log=TRUE))
  }
  est.para <- optim(parastart, optimfun, method=method, control = list(maxit = 2e+7, trace = trace, fnscale=-1,...), hessian=hessian)
  if(hessian){
    return(list(est=est.para$par, stderr=sqrt(diag(solve(-est.para$hessian))), hess=est.para$hessian, LL=est.para$value))
  }else{
    return(list(est=est.para$par, LL= est.para$value ))
  }
  
}

fit.hier.univ.hist.normS.fDF <- function(symbol.data, parastart, df, method="Nelder-Mead", hessian=TRUE, trace=0,...){
  m <- length(symbol.data) # number of symbols
  optimfun <- function(para){
    return(llik.hier.univ.hist.normS(symbol.data=symbol.data, Theta= para[1:(m*2)], varphi=c(para[(m*2 +1):(m*2 + 4)], df,para[(m*2 +5):(m*2 + 6)]), log=TRUE))
  }
  est.para <- optim(parastart, optimfun, method=method, control = list(maxit = 2e+7, trace = trace, fnscale=-1,...), hessian=hessian)
  if(hessian){
    return(list(est=est.para$par, stderr=sqrt(diag(solve(-est.para$hessian))), hess=est.para$hessian, LL=est.para$value))
  }else{
    return(list(est=est.para$par, LL= est.para$value ))
  }
  
}


###################################################################
########## SKEW-NORMAL DISTRIBUTION ###############################
###################################################################

########## Inclusion of the distribution of the scale #############
########## Inclusion of the distribution of the skewness ##########

dens.hier.univ.hist.SKnormSS <- function(symbol, theta, parametrisation="dp", varphi, log){
  
  # "symbol" is a list containing elements:
  #        - 'counts' 
  #        - 'breaks'
  #        - 'risk.lvl' constaining a value from 1 to 35 correspond to the risk level A1 to G5
  # "theta" a vector of parameters: location, scale and skewness, length is 3 
  # "varphi" is a vector of parameters for the distribution of the mean and variance
  #        - The location is T distributed with:
  #             -- mean as function of the risk level, "a risk^2 + b risk + c" and "d risk^2 + e risk + f";
  #             -- variance;
  #             -- degree of freedom "nu";
  #        - The scale matrix is Inverse-Gamma distributed  with:
  #             -- degree of freedom "v";
  #             -- a scale matrix "S";
  #        - The skewness is Normally distributed with:
  #             -- mean "mu.sk";
  #             -- variance "sig.sk";
  # "log" to decide if the density or the log-density is returned
  
  if(length(varphi) != 9){ stop("Wrong length of parameter varphi") }
  if(!is.integer(symbol$risk.lvl) || symbol$risk.lvl < 1 || symbol$risk.lvl > 35){ stop("risk.lvl must be an integer between 1 and 35") }
  
  a = varphi[1]; b = varphi[2]; c = varphi[3]; 
  ohm = varphi[4]; nu = varphi[5];  
  shape = varphi[6]; scale = varphi[7];
  mu.sk = varphi[8]; sig.sk = varphi[9];
  
  if(ohm <=0 || nu <= 0 || shape <= 0 || scale <= 0 || sig.sk <= 0){if(log){return(-1e300)}else{return(1e-300)}}
  
  if(length(theta) != 3){stop("Wrong length of parameters")}
  if(theta[2] <=0){if(log){return(-1e300)}else{return(1e-300)}}
  
  if(parametrisation=="dp") dp <-theta
  if(parametrisation=="cp"){
    if(abs(theta[3]) >= (4-pi)/2 * (2 / (pi-2))^(3/2) ){if(log){return(-1e300)}else{return(1e-300)}}
    dp <- cp2dp(c(theta[1], sqrt(theta[2]), theta[3]) , "SN")
    dp[2] <- dp[2]^2
  } 
  
  d1 <- dsymbolic1(counts=symbol$counts, breaks=symbol$breaks, distribution="Sk-Normal", param=dp, log=log)
  d2 <- dt(x=c(theta[1]- a * symbol$risk.lvl^2 - b * symbol$risk.lvl - c)/sqrt(ohm), df=nu, log=log )
  if(log) d3 <- log(ohm)/2 else d3 <- sqrt(ohm)
  
  if(d2==-Inf) d2 <- -1e300
  d4 <- dinvgamma(x=theta[2], shape=shape, scale=scale)
  d5 <- dnorm(x=theta[3], mean=mu.sk, sd=sqrt(sig.sk), log=log)
  
  if(log) return(d1 + d2 - d3 + log(d4) + d5) else return(d1 * d2 / d3 * d4 * d5)
  
}


llik.hier.univ.hist.SKnormSS <- function(symbol.data, Theta, parametrisation="dp", varphi, log=TRUE, vectorial=FALSE, parallel=FALSE){
  # symbol.data is the symbolic dataset, a list of symbols as defined in dens.hier.biv.int.norm
  # If vectorial=TRUE then a vector of each log-likelihood evaluation is returned, otherwise the sum is returned
  #
  # Theta is a vector of all the theta_i, i being the group index
  # so for the Normal model each theta is of length 5 and thus Theta is of length m*5
  
  #cat("Params: ", Theta, "\n")
  m <- length(symbol.data) # number of symbols
  
  if(length(Theta) != m*3){stop("Wrong lentgh of Theta vector")}
  
  if(parallel){
    llik <- foreach(i=1:m, .combine = c, .export = c("dens.hier.univ.hist.SKnormSS", "dsymbolic1", "psn", "dinvgamma")) %dopar% {
      dens.hier.univ.hist.SKnormSS(symbol=symbol.data[[i]], theta=Theta[((i-1)*3+1):(i*3)], parametrisation=parametrisation, varphi=varphi, log=log)
    }
  }else{
    llik <- vector("numeric",length=m)
    for( j in 1:m){
      llik[j] <- dens.hier.univ.hist.SKnormSS(symbol=symbol.data[[j]], theta=Theta[((j-1)*3+1):(j*3)], parametrisation=parametrisation, varphi=varphi, log=log)
    }
  }
  
  if(vectorial){
    return(llik)
  }else{
    if(log) return(sum(llik)) else return(prod(llik))
  }
  
}  

fit.hier.univ.hist.SKnormSS <- function(symbol.data, parastart, parametrisation="dp", parallel=FALSE, method="Nelder-Mead", hessian=TRUE, trace=0,...){
  m <- length(symbol.data) # number of symbols
  optimfun <- function(para){
    return(llik.hier.univ.hist.SKnormSS(symbol.data=symbol.data, Theta= para[1:(m*3)], parametrisation=parametrisation, varphi=para[(m*3 +1):(m*3 + 9)], log=TRUE, parallel=parallel))
  }
  est.para <- optim(parastart, optimfun, method=method, control = list(maxit = 2e+7, trace = trace, fnscale=-1,...), hessian=hessian)
  if(hessian){
    return(list(est=est.para$par, stderr=sqrt(diag(solve(-est.para$hessian))), hess=est.para$hessian, LL=est.para$value))
  }else{
    return(list(est=est.para$par, LL= est.para$value ))
  }
  
}

fit.hier.univ.hist.SKnormSS.fDF <- function(symbol.data, parastart, parametrisation="dp", df, parallel=FALSE, method="Nelder-Mead", hessian=TRUE, trace=0,...){
  m <- length(symbol.data) # number of symbols
  optimfun <- function(para){
    return(llik.hier.univ.hist.SKnormSS(symbol.data=symbol.data, Theta= para[1:(m*3)], parametrisation=parametrisation, varphi=c(para[(m*3 +1):(m*3 + 4)], df, para[(m*3 +5):(m*3 + 8)]), log=TRUE, parallel=parallel))
  }
  est.para <- optim(parastart, optimfun, method=method, control = list(maxit = 2e+7, trace = trace, fnscale=-1,...), hessian=hessian)
  if(hessian){
    return(list(est=est.para$par, stderr=sqrt(diag(solve(-est.para$hessian))), hess=est.para$hessian, LL=est.para$value))
  }else{
    return(list(est=est.para$par, LL= est.para$value ))
  }
  
}
