###
###
### Functions for the hierarchical model
###		for univariate interval symbols
###		assuming Normal and Skew-Normal distributions
###

### Multinomial coefficient is not including as it is not needed in the parameter estimation

###
###
###
### density univariate interval for both normal and skew-normal distributions
###
###
###

dens.univ.int <- function(symbol, param, distribution="Normal", parametrisation="dp", log=FALSE){
  
  # "symbol" is a vector of length 5: min, max, n, l-th and u-th order statistics;
  # "param" is of length 2 if distribution="Normal";
  #               length 3 if distribution="Sk-Normal";
  
  if( length(symbol) != 5 ){stop("Wrong symbol length") }
  if(is.null(symbols) || is.null(param) || is.null(distribution)){stop("Missing input") }
  
  if(log){p <- 0}else{p <- 1}
  
  ## For underlying normal distribution
  
  if(distribution=='Normal'){
    
    if(length(param) != 2){stop("Wrong length of parameters")}
    if(param[2] <=0 ){if(log){return(log(-1e320))}else{return(1e-320)}}
    
    p1 <- pnorm(symbol[1], mean=param[1], sd=sqrt(param[2]))
    p2 <- pnorm(symbol[2], mean=param[1], sd=sqrt(param[2]))
    
    Dens <- dnorm(symbol[1:2], mean=param[1], sd=sqrt(param[2]), log=log )
    
  } ## END: if distribution=="Normal"
  
  ## For underlying skew-normal distribution
  
  if(distribution=='Sk-Normal'){
    
    if(length(param) != 3){stop("Wrong length of parameters")}
    if(param[2] <=0){if(log){return(-1e300)}else{return(1e-300)}}
    
    if(parametrisation=="dp") dp <-c(param[1], sqrt(param[2]), param[3])
    if(parametrisation=="cp"){
      if(abs(param[3]) >= (4-pi)/2 * (2 / (pi-2))^(3/2) ){if(log){return(-1e300)}else{return(1e-300)}}
      dp <- cp2dp(c(param[1], sqrt(param[2]), param[3]) , "SN")
    } 
    
    p1 <- psn(symbol[1], dp=dp)
    p2 <- psn(symbol[2], dp=dp)
    Dens <- dsn(symbol[1:2], dp=dp, log=log )
    
  } ## END: if distribution=="Sk-Normal" 
  
  if(p1 == 0){ p1 <- 1e-300}
  if(p2 == 1){ np2 <- 1e-300  }else np2 <- 1-p2
  
  if(log){
    p <- p + (symbol[4]-1) * log( p1 ) + (symbol[5]-symbol[4]-1) * log(p2-p1) + (symbol[3]-symbol[5]) * log(np2) + sum(Dens)
  }else{
    p <- p * p1^(symbol[4]-1) * ( p2-p1 )^(symbol[5]-symbol[4]-1) * (1-p2)^(symbol[3]-symbol[5]) * prod(Dens)
  } 
  
  return(p)	
}


###
###
###
### FUNCTIONS FOR HIERARCHICAL MODEL
###
###
###


###################################################################
########## NORMAL DISTRIBUTION ####################################
###################################################################

########## Inclusion of the distribution of the scale #############

dens.hier.univ.int.normS <- function(symbol, theta, varphi, log){
  
  # "symbol" is a list containing elements:
  #        - 'interv' a vector of length 3 giving the min, max and n; 
  #        - 'risk.lvl' constaining a value from 1 to 35 correspond to the risk level A1 to G5
  # "theta" a vector of parameters: location and scale (var), length is 2 
  # "varphi" is a vector of parameters for the distribution of the mean and variance
  #        - The location is T distributed with:
  #             -- mean as function of the risk level, "a risk^2 + b risk + c";
  #             -- variance "ohm";
  #             -- degree of freedom "nu";
  #        - The scale is Inverse-Gamma distributed  with:
  #             -- shape;
  #             -- scale;
  # "log" to decide if the density or the log-density is returned
  
  if(length(varphi) != 7){ stop("Wrong length of parameter varphi") }
  if(!is.integer(symbol$risk.lvl) || symbol$risk.lvl < 1 || symbol$risk.lvl > 35){ stop("risk.lvl must be an integer between 1 and 35") }
  
  a = varphi[1]; b = varphi[2]; c = varphi[3]; 
  ohm = varphi[4]; nu = varphi[5];  
  shape = varphi[6]; scale = varphi[7];
  
  if(ohm <=0 || nu <= 0 || shape <= 0 || scale <= 0){if(log){return(-1e300)}else{return(1e-300)}}
  
  d1 <- dens.univ.int(symbol=symbol$interv, param=theta, distribution="Normal", log=log)
  d2 <- dt(x=c(theta[1]- a * symbol$risk.lvl^2 - b * symbol$risk.lvl - c)/sqrt(ohm), df=nu, log=log )
  if(log) d3 <- log(ohm)/2 else d3 <- sqrt(ohm)
  
  if(d2==-Inf) d2 <- -1e300
  d4 <- dinvgamma(x=theta[2], shape=shape, scale=scale)
  
  if(log) return(d1 + d2 - d3 + log(d4)) else return(d1 * d2 / d3 * d4)
  
}


llik.hier.univ.int.normS <- function(symbol.data, Theta, varphi, log=TRUE, vectorial=FALSE){
  # symbol.data is the symbolic dataset, a list of symbols as defined in dens.hier.biv.int.norm
  # If vectorial=TRUE then a vector of each log-likelihood evaluation is returned, otherwise the sum is returned
  #
  # Theta is a vector of all the theta_i, i being the group index
  # so for the Normal model each theta is of length 5 and thus Theta is of length m*2
  
  #cat("Params: ", Theta, "\n")
  m <- length(symbol.data) # number of symbols
  llik <- vector("numeric",length=m)
  
  if(length(Theta) != m*2){stop("Wrong lentgh of Theta vector")}
  
  for( j in 1:m){
    llik[j] <- dens.hier.univ.int.normS(symbol=symbol.data[[j]], theta=Theta[((j-1)*2+1):(j*2)], varphi=varphi, log=log)
  }
  
  if(vectorial){
    return(llik)
  }else{
    if(log) return(sum(llik)) else return(prod(llik))
  }
  
}  


fit.hier.univ.int.normS <- function(symbol.data, parastart, method="Nelder-Mead", hessian=TRUE, trace=0,...){
  m <- length(symbol.data) # number of symbols
  optimfun <- function(para){
    return(llik.hier.univ.int.normS(symbol.data=symbol.data, Theta= para[1:(m*2)], varphi=para[(m*2 +1):(m*2 + 7)], log=TRUE))
  }
  est.para <- optim(parastart, optimfun, method=method, control = list(maxit = 2e+7, trace = trace, fnscale=-1,...), hessian=hessian)
  if(hessian){
    return(list(est=est.para$par, stderr=sqrt(diag(solve(-est.para$hessian))), hess=est.para$hessian, LL=est.para$value))
  }else{
    return(list(est=est.para$par, LL= est.para$value ))
  }
  
}

fit.hier.univ.int.normS.fDF <- function(symbol.data, parastart, df, method="Nelder-Mead", hessian=TRUE, trace=0,...){
  m <- length(symbol.data) # number of symbols
  optimfun <- function(para){
    return(llik.hier.univ.int.normS(symbol.data=symbol.data, Theta= para[1:(m*2)], varphi=c(para[(m*2 +1):(m*2 + 4)], df, para[(m*2 +5):(m*2 + 6)]), log=TRUE))
  }
  est.para <- optim(parastart, optimfun, method=method, control = list(maxit = 2e+7, trace = trace, fnscale=-1,...), hessian=hessian)
  if(hessian){
    return(list(est=est.para$par, stderr=sqrt(diag(solve(-est.para$hessian))), hess=est.para$hessian, LL=est.para$value))
  }else{
    return(list(est=est.para$par, LL= est.para$value ))
  }
  
}

###################################################################
########## SKEW-NORMAL DISTRIBUTION ###############################
###################################################################

########## Inclusion of the distribution of the scale #############
########## Inclusion of the distribution of the skewness ##########

dens.hier.univ.int.SKnormSS <- function(symbol, theta, parametrisation="dp",varphi, log){
  
  # "symbol" is a list containing elements:
  #        - 'interv' a vector of length 3 giving the min, max and n; 
  #        - 'risk.lvl' constaining a value from 1 to 35 correspond to the risk level A1 to G5
  # "theta" a vector of parameters: location and scale (var), length is 3 
  # "varphi" is a vector of parameters for the distribution of the mean and variance
  #        - The location is T distributed with:
  #             -- mean as function of the risk level, "a risk^2 + b risk + c";
  #             -- variance "ohm";
  #             -- degree of freedom "nu";
  #        - The scale is Inverse-Gamma distributed  with:
  #             -- shape;
  #             -- scale;
  #        - The skewness is Normally distributed with:
  #             -- mean "mu.sk";
  #             -- variance "sig.sk";
  # "log" to decide if the density or the log-density is returned
  
  if(length(varphi) != 9){ stop("Wrong length of parameter varphi") }
  if(!is.integer(symbol$risk.lvl) || symbol$risk.lvl < 1 || symbol$risk.lvl > 35){ stop("risk.lvl must be an integer between 1 and 35") }
  
  a = varphi[1]; b = varphi[2]; c = varphi[3]; 
  ohm = varphi[4]; nu = varphi[5];  
  shape = varphi[6]; scale = varphi[7];
  mu.sk = varphi[8]; sig.sk = varphi[9];
  
  if(ohm <=0 || nu <= 0 || shape <= 0 || scale <= 0 || sig.sk <= 0){if(log){return(-1e300)}else{return(1e-300)}}
  
  if(length(theta) != 3){stop("Wrong length of parameters")}
  if(theta[2] <=0){if(log){return(-1e300)}else{return(1e-300)}}
  
  if(parametrisation=="dp") dp <-theta
  if(parametrisation=="cp"){
    if(abs(theta[3]) >= (4-pi)/2 * (2 / (pi-2))^(3/2) ){if(log){return(-1e300)}else{return(1e-300)}}
    dp <- cp2dp(c(theta[1], sqrt(theta[2]), theta[3]) , "SN")
    dp[2] <- dp[2]^2
  } 
  
  d1 <- dens.univ.int(symbol=symbol$interv, param=dp, distribution="Sk-Normal", log=log)
  d2 <- dt(x=c(theta[1]- a * symbol$risk.lvl^2 - b * symbol$risk.lvl - c)/sqrt(ohm), df=nu, log=log )
  if(log) d3 <- log(ohm)/2 else d3 <- sqrt(ohm)
  
  if(d2==-Inf) d2 <- -1e300
  d4 <- dinvgamma(x=theta[2], shape=shape, scale=scale)
  d5 <- dnorm(x=theta[3], mean=mu.sk, sd=sqrt(sig.sk), log=log)
  
  if(log) return(d1 + d2 - d3 + log(d4) + d5) else return(d1 * d2 / d3 * d4 * d5)
  
}


llik.hier.univ.int.SKnormSS <- function(symbol.data, Theta, parametrisation="dp", varphi, log=TRUE, vectorial=FALSE){
  # symbol.data is the symbolic dataset, a list of symbols as defined in dens.hier.biv.int.norm
  # If vectorial=TRUE then a vector of each log-likelihood evaluation is returned, otherwise the sum is returned
  #
  # Theta is a vector of all the theta_i, i being the group index
  # so for the Normal model each theta is of length 3 and thus Theta is of length m*3
  
  #cat("Params: ", Theta, "\n")
  m <- length(symbol.data) # number of symbols
  llik <- vector("numeric",length=m)
  
  if(length(Theta) != m*3){stop("Wrong lentgh of Theta vector")}
  
  for( j in 1:m){
    llik[j] <- dens.hier.univ.int.SKnormSS(symbol=symbol.data[[j]], theta=Theta[((j-1)*3+1):(j*3)], parametrisation=parametrisation, varphi=varphi, log=log)
  }
  
  if(vectorial){
    return(llik)
  }else{
    if(log) return(sum(llik)) else return(prod(llik))
  }
  
}  


fit.hier.univ.int.SKnormSS <- function(symbol.data, parastart, parametrisation="dp", method="Nelder-Mead", hessian=TRUE, trace=0,...){
  m <- length(symbol.data) # number of symbols
  optimfun <- function(para){
    return(llik.hier.univ.int.SKnormSS(symbol.data=symbol.data, Theta= para[1:(m*3)], parametrisation=parametrisation, varphi=para[(m*3 +1):(m*3 + 9)], log=TRUE))
  }
  est.para <- optim(parastart, optimfun, method=method, control = list(maxit = 2e+7, trace = trace, fnscale=-1,...), hessian=hessian)
  if(hessian){
    return(list(est=est.para$par, stderr=sqrt(diag(solve(-est.para$hessian))), hess=est.para$hessian, LL=est.para$value))
  }else{
    return(list(est=est.para$par, LL= est.para$value ))
  }
  
}

fit.hier.univ.int.SKnormSS.fDF <- function(symbol.data, parastart, parametrisation="dp", df, method="Nelder-Mead", hessian=TRUE, trace=0,...){
  m <- length(symbol.data) # number of symbols
  optimfun <- function(para){
    return(llik.hier.univ.int.SKnormSS(symbol.data=symbol.data, Theta= para[1:(m*3)], parametrisation=parametrisation, varphi=c(para[(m*3 +1):(m*3 + 4)], df,para[(m*3 +5):(m*3 + 8)]), log=TRUE))
  }
  est.para <- optim(parastart, optimfun, method=method, control = list(maxit = 2e+7, trace = trace, fnscale=-1,...), hessian=hessian)
  if(hessian){
    return(list(est=est.para$par, stderr=sqrt(diag(solve(-est.para$hessian))), hess=est.para$hessian, LL=est.para$value))
  }else{
    return(list(est=est.para$par, LL= est.para$value ))
  }
  
}


