library(gplots)
library(mvtnorm)
library(sn)
library(MCMCpack)

rm(list=ls())
source("functions_billard.R")
source("functions_uni_hist_N_SN.R")
source("functions_uni_int_N_SN.R")
source("functions_uni_N_SN.R")
loanDF <- read.csv("loan.csv")

attach(loanDF)
risks <-c(levels(sub_grade))

## focus on (log) annual income within sub_grade
loanDF1 <- loanDF[,names(loanDF) %in% c("annual_inc", "sub_grade")]

## remove annual incomes that are NA or 0 
## there are 887373 observations, 2 variables 
loanDF1 <- na.omit(loanDF1, cols=c("annual_inc", "int_rate"))
loanDF1 <- loanDF1[-c( which(loanDF1$annual_inc==0) ),]

## Variables on the log scale
loanDF1$annual_inc_ln = log(loanDF1$annual_inc)

###
### EXPLORATORY DATA ANALYSIS 
### + Creation of symbols
### + Likelihood Ratio test for Skewness
###

class.data <- class.data.sk <- symbols.hist <- symbols.int <- symbols.B.int <- symbols.B.hist <- list(length=length(risks))
analysis.logIncome <- matrix(ncol=5, nrow=length(risks))
colnames(analysis.logIncome) <- c("Risk", "n", "alpha", "pval", "model")
D <- Ns <- M <- M.int <- M.hist <- S <- S.int <- S.hist <- vector(length=length(risks))

for( i in 1:length(risks)){
  
  ## Annual income by risk group
  loan_sub = loanDF1[which(loanDF1$sub_grade==risks[i]) , ]
  n <- nrow(loan_sub)
  Ns[i] <- n
  
  tmp <- loan_sub$annual_inc_ln
  M[i] <- mean(tmp)
  S[i] <- sd(tmp)
  
  ## Classical data
  class.data[[i]] <- list( data=c(M[i],S[i],Ns[i]), risk.lvl=i )
  class.data.sk[[i]] <- list( data=tmp, risk.lvl=i )
  
  ## Symbolic data (histograms)   
  x.breaks <- seq(from=min(tmp)-1e-10, to=max(tmp)+1e-10,  length=6 ) # 5 bin histogram   
  symbols.hist[[i]] <-  hist(tmp, breaks=x.breaks, plot=FALSE)
  symbols.hist[[i]]$risk.lvl <- i
  A <- head(symbols.hist[[i]]$breaks,-1)
  B <- tail(symbols.hist[[i]]$breaks,-1)
  M.hist[i] <- sum((A+B) * symbols.hist[[i]]$counts)  /(2*sum(symbols.hist[[i]]$counts))
  S.hist[i] <- sum(symbols.hist[[i]]$counts * (A^2 + A*B + B^2) )/(3*n) - (M.hist[i])^2
  symbols.B.hist[[i]] <- list( val = c(M.hist[i], S.hist[i]), risk.lvl = i)
  
  ## Symbolic data (intervals)
  symbols.int[[i]] <- list(interv = c(min(tmp), max(tmp), n, 1, n), risk.lvl = i)
  M.int[i] <- (symbols.int[[i]]$interv[1] + symbols.int[[i]]$interv[2])/2
  S.int[i] <- (symbols.int[[i]]$interv[2]-symbols.int[[i]]$interv[1])^2/12
  symbols.B.int[[i]] <- list(val = c(M.int[i], S.int[i]), risk.lvl = i)
  
  ## Fit skew-normal distribution
  snorm.fit <- selm.fit(x= matrix(rep(1, n), n, 1),  y = tmp, family = "SN", fixed.param = list(), selm.control="MLE")
  
  ## Fit normal distribution
  norm.fit <- selm.fit(x= matrix(rep(1, n), n, 1),  y = tmp, family = "SN", fixed.param = list(alpha=0), selm.control="MLE")
 
  ## Likelihood ratio test: (H0) alpha = 0 vs. (H1) alpha != 0
  D[i] = 2 * (snorm.fit$logL - norm.fit$logL)
  pval <- 1-pchisq(D[i],df=1)
  if(pval < 0.05) model <- "SN" else model <- "N"
  
  analysis.logIncome[i,] <- c(risks[i], n, snorm.fit$param$dp[3], pval, model)
    
}

####
####
#### ANALYSIS OF LOG-INCOME
####
####


###
###
###
###
### LE RADEMACHER & BILLARD (2011) METHOD
###
###
###
###

### Using Intervals

BillardInt.fDF <- c(0,0,11,1, 1,1 )
LL.BI.fDF <- LL.diffBI.fDF <- llik.billard(symbol.data = symbols.int, type="int", param=c(BillardInt.fDF[1:4], 3, BillardInt.fDF[5:6]))
resBI.fDF <- c(BillardInt.fDF, LL.BI.fDF, LL.diffBI.fDF)

TimeBI.fDF <- system.time(
  while(abs(LL.diffBI.fDF)>0.001){
    
    estBI.fDF <- fit.billard.fDF(symbol.data=symbols.int, type="int", parastart=BillardInt.fDF, df=3, method="BFGS", hessian=FALSE, trace=2)
    LL.diffBI.fDF <- estBI.fDF$LL - LL.BI.fDF
    LL.BI.fDF <- estBI.fDF$LL
    BillardInt.fDF <- estBI.fDF$est
    
    resBI.fDF <- rbind(resBI.fDF, c(BillardInt.fDF, LL.BI.fDF, LL.diffBI.fDF))

    estBI.fDF <- fit.billard.fDF(symbol.data=symbols.int, type="int", parastart=BillardInt.fDF, df=3, method="Nelder-Mead", hessian=FALSE, trace=2)
    LL.diffBI.fDF <- estBI.fDF$LL - LL.BI.fDF
    LL.BI.fDF <- estBI.fDF$LL
    BillardInt.fDF <- estBI.fDF$est
    
    resBI.fDF <- rbind(resBI.fDF, c(BillardInt.fDF, LL.BI.fDF, LL.diffBI.fDF))
    
  }
)

### Using histograms

BillardHist.fDF <- c(0,0,11,1, 1,1 )
LL.BH.fDF <- LL.diffBH.fDF <- llik.billard(symbol.data = symbols.hist, type="hist", param=c(BillardHist.fDF[1:4], 3, BillardHist.fDF[5:6]))
resBH.fDF <- c(BillardHist.fDF, LL.BH.fDF, LL.diffBH.fDF)

TimeBH.fDF <- system.time(
  while(abs(LL.diffBH.fDF)>0.001){
    
    estBH.fDF <- fit.billard.fDF(symbol.data=symbols.hist, type="hist", parastart=BillardHist.fDF, df=3, method="BFGS", hessian=FALSE, trace=2)
    LL.diffBH.fDF <- estBH.fDF$LL - LL.BH.fDF
    LL.BH.fDF <- estBH.fDF$LL
    BillardHist.fDF <- estBH.fDF$est
    
    resBH.fDF <- rbind(resBH.fDF, c(BillardHist.fDF, LL.BH.fDF, LL.diffBH.fDF))
    
    estBH.fDF <- fit.billard.fDF(symbol.data=symbols.hist, type="hist", parastart=BillardHist.fDF, df=3, method="Nelder-Mead", hessian=FALSE, trace=2)
    LL.diffBH.fDF <- estBH.fDF$LL - LL.BH.fDF
    LL.BH.fDF <- estBH.fDF$LL
    BillardHist.fDF <- estBH.fDF$est
    
    resBH.fDF <- rbind(resBH.fDF, c(BillardHist.fDF, LL.BH.fDF, LL.diffBH.fDF))
    
  }
)


###
###
###
###
### NORMAL DISTRIBUTION
###
###
###
###

##########################################################################################
########## t distribution for the mean              ######################################
########## inverse-Gamma distribution for the scale ######################################
##########################################################################################

### Histograms

estimS.fDF <- c(rep(c(11.5, 1),35), 0,0,11,1, 1,1 )
LLS.fDF <- LL.diffS.fDF <- llik.hier.univ.hist.normS(symbol.data = symbols.hist, Theta = estimS.fDF[1:70] , varphi = c(estimS.fDF[71:74], 3, estimS.fDF[75:76]))
resS.fDF <- c(estimS.fDF, LLS.fDF, LL.diffS.fDF)

TimeS.fDF <- system.time(
  while(abs(LL.diffS.fDF)>0.001){
    
    estS.fDF <- fit.hier.univ.hist.normS.fDF(symbol.data = symbols.hist, parastart = estimS.fDF, df=3, method="BFGS", hessian=FALSE, trace=2)
    LL.diffS.fDF <- estS.fDF$LL - LLS.fDF
    LLS.fDF <- estS.fDF$LL
    estimS.fDF <- estS.fDF$est
    
    resS.fDF <- rbind(resS.fDF, c(estimS.fDF, LLS.fDF, LL.diffS.fDF))
    
    estS.fDF <- fit.hier.univ.hist.normS.fDF(symbol.data = symbols.hist, parastart = estimS.fDF, df=3, method="Nelder-Mead", hessian=FALSE, trace=2)
    LL.diffS.fDF <- estS.fDF$LL - LLS.fDF
    LLS.fDF <- estS.fDF$LL
    estimS.fDF <- estS.fDF$est
    
    resS.fDF <- rbind(resS.fDF, c(estimS.fDF, LLS.fDF, LL.diffS.fDF))
    
  }
)

### Intervals

estimIntS.fDF <- c(rep(c(11.5, 1),35), 0,0,11,1, 1,1 )
LLIntS.fDF <- LL.diffIntS.fDF <- llik.hier.univ.int.normS(symbol.data = symbols.int, Theta = estimIntS.fDF[1:70] , varphi = c(estimIntS.fDF[71:74], 3, estimIntS.fDF[75:76]))
resIntS.fDF <- c(estimIntS.fDF, LLIntS.fDF, LL.diffIntS.fDF)

TimeIntS.fDF <- system.time(
  while(abs(LL.diffIntS.fDF)>0.001){
    
    estIntS.fDF <- fit.hier.univ.int.normS.fDF(symbol.data = symbols.int, parastart = estimIntS.fDF, df=3, method="BFGS", hessian=FALSE, trace=2)
    LL.diffIntS.fDF <- estIntS.fDF$LL - LLIntS.fDF
    LLIntS.fDF <- estIntS.fDF$LL
    estimIntS.fDF <- estIntS.fDF$est
    
    resIntS.fDF <- rbind(resIntS.fDF, c(estimIntS.fDF, LLIntS.fDF, LL.diffIntS.fDF))
    
    estIntS.fDF <- fit.hier.univ.int.normS.fDF(symbol.data = symbols.int, parastart = estimIntS.fDF, df=3, method="Nelder-Mead", hessian=FALSE, trace=2)
    LL.diffIntS.fDF <- estIntS.fDF$LL - LLIntS.fDF
    LLIntS.fDF <- estIntS.fDF$LL
    estimIntS.fDF <- estIntS.fDF$est
    
    resIntS.fDF <- rbind(resIntS.fDF, c(estimIntS.fDF, LLIntS.fDF, LL.diffIntS.fDF))
     
  }
)

### Classical

estimS.fDF.c <- c(rep(c(11.5, 1),35), 0,0,11,1, 1,1 )
LLS.fDF.c <- LL.diffS.fDF.c <- llik.hier.univ.normS(data = class.data, Theta = estimS.fDF.c[1:70] , varphi = c(estimS.fDF.c[71:74],3,estimS.fDF.c[75:76]))
resS.fDF.c <- c(estimS.fDF.c, LLS.fDF.c, LL.diffS.fDF.c)

TimeS.fDF.c <- system.time(
  while(abs(LL.diffS.fDF.c)>0.001){
    
    estS.fDF.c <- fit.hier.univ.normS.fDF(data = class.data, parastart = estimS.fDF.c, df=3, method="BFGS", hessian=FALSE, trace=2)
    LL.diffS.fDF.c <- estS.fDF.c$LL - LLS.fDF.c
    LLS.fDF.c <- estS.fDF.c$LL
    estimS.fDF.c <- estS.fDF.c$est
    
    resS.fDF.c <- rbind(resS.fDF.c, c(estimS.fDF.c, LLS.fDF.c, LL.diffS.fDF.c))
    
    estS.fDF.c <- fit.hier.univ.normS.fDF(data = class.data, parastart = estimS.fDF.c, df=3, method="Nelder-Mead", hessian=FALSE, trace=2)
    LL.diffS.fDF.c <- estS.fDF.c$LL - LLS.fDF.c
    LLS.fDF.c <- estS.fDF.c$LL
    estimS.fDF.c <- estS.fDF.c$est
    
    resS.fDF.c <- rbind(resS.fDF.c, c(estimS.fDF.c, LLS.fDF.c, LL.diffS.fDF.c))
    
  }
)


###
###
###
###
### SKEW NORMAL DISTRIBUTION
###
###
###
###


##########################################################################################
########## t distribution for the mean              ######################################
########## inverse-Gamma distribution for the scale ######################################
########## normal distribution for the skewness     ######################################
##########################################################################################

### Histograms

estimSkSS.cp.fDF <- c(rep(c(11.5, 1, 0.1),35), 0,0,11,1, 1,1, 1,2 )
LLSkSS.cp.fDF <- LL.diffSkSS.cp.fDF <- llik.hier.univ.hist.SKnormSS(symbol.data = symbols.hist, Theta = estimSkSS.cp.fDF[1:105], parametrisation="cp" , varphi = c(estimSkSS.cp.fDF[106:109],3,estimSkSS.cp.fDF[110:113]))
resSkSS.cp.fDF <- c(estimSkSS.cp.fDF, LLSkSS.cp.fDF, LL.diffSkSS.cp.fDF)

TimeSkSS.cp.fDF <- system.time(
  while(abs(LL.diffSkSS.cp.fDF)>0.001){
    
    estSkSS.cp.fDF <- fit.hier.univ.hist.SKnormSS.fDF(symbol.data = symbols.hist, parastart = estimSkSS.cp.fDF, parametrisation="cp", df=3, method="BFGS", hessian=FALSE, trace=2)
    LL.diffSkSS.cp.fDF <- estSkSS.cp.fDF$LL - LLSkSS.cp.fDF
    LLSkSS.cp.fDF <- estSkSS.cp.fDF$LL
    estimSkSS.cp.fDF <- estSkSS.cp.fDF$est
    
    resSkSS.cp.fDF <- rbind(resSkSS.cp.fDF, c(estimSkSS.cp.fDF, LLSkSS.cp.fDF, LL.diffSkSS.cp.fDF))
    
    estSkSS.cp.fDF <- fit.hier.univ.hist.SKnormSS.fDF(symbol.data = symbols.hist, parastart = estimSkSS.cp.fDF, parametrisation="cp", df=3, method="Nelder-Mead", hessian=FALSE, trace=2)
    LL.diffSkSS.cp.fDF <- estSkSS.cp.fDF$LL - LLSkSS.cp.fDF
    LLSkSS.cp.fDF <- estSkSS.cp.fDF$LL
    estimSkSS.cp.fDF <- estSkSS.cp.fDF$est
    
    resSkSS.cp.fDF <- rbind(resSkSS.cp.fDF, c(estimSkSS.cp.fDF, LLSkSS.cp.fDF, LL.diffSkSS.cp.fDF))
    
  }
)

### Intervals

estimIntSkSS.cp.fDF <- c(rep(c(11.5, 1, 0.1),35), 0,0,11,1, 1,1, 1,2 )
LLIntSkSS.cp.fDF <- LL.diffIntSkSS.cp.fDF <- llik.hier.univ.int.SKnormSS(symbol.data = symbols.int, Theta = estimIntSkSS.cp.fDF[1:105], parametrisation="cp" , varphi = c(estimIntSkSS.cp.fDF[106:109],3,estimIntSkSS.cp.fDF[110:113]))
resIntSkSS.cp.fDF <- c(estimIntSkSS.cp.fDF, LLIntSkSS.cp.fDF, LL.diffIntSkSS.cp.fDF)

TimeIntSkSS.cp.fDF <- system.time(
  while(abs(LL.diffIntSkSS.cp.fDF)>0.001){
    
    estIntSkSS.cp.fDF <- fit.hier.univ.int.SKnormSS.fDF(symbol.data = symbols.int, parastart = estimIntSkSS.cp.fDF, parametrisation="cp", df=3, method="BFGS", hessian=FALSE, trace=2)
    LL.diffIntSkSS.cp.fDF <- estIntSkSS.cp.fDF$LL - LLIntSkSS.cp.fDF
    LLIntSkSS.cp.fDF <- estIntSkSS.cp.fDF$LL
    estimIntSkSS.cp.fDF <- estIntSkSS.cp.fDF$est
    
    resIntSkSS.cp.fDF <- rbind(resIntSkSS.cp.fDF, c(estimIntSkSS.cp.fDF, LLIntSkSS.cp.fDF, LL.diffIntSkSS.cp.fDF))
    
    estIntSkSS.cp.fDF <- fit.hier.univ.int.SKnormSS.fDF(symbol.data = symbols.int, parastart = estimIntSkSS.cp.fDF, parametrisation="cp", df=3, method="Nelder-Mead", hessian=FALSE, trace=2)
    LL.diffIntSkSS.cp.fDF <- estIntSkSS.cp.fDF$LL - LLIntSkSS.cp.fDF
    LLIntSkSS.cp.fDF <- estIntSkSS.cp.fDF$LL
    estimIntSkSS.cp.fDF <- estIntSkSS.cp.fDF$est
    
    resIntSkSS.cp.fDF <- rbind(resIntSkSS.cp.fDF, c(estimIntSkSS.cp.fDF, LLIntSkSS.cp.fDF, LL.diffIntSkSS.cp.fDF))
    
  }
)

### Classical

estimSkSS.cp.fDF.c <- c(rep(c(11.5, 1, 0.1),35), 0,0,11,1, 1,1, 1,2 )
LLSkSS.cp.fDF.c <- LL.diffSkSS.cp.fDF.c <- llik.hier.univ.SKnormSS(data = class.data.sk, Theta = estimSkSS.cp.fDF.c[1:105] , parametrisation="cp", varphi = c(estimSkSS.cp.fDF.c[106:109],3,estimSkSS.cp.fDF.c[110:113]))
resSkSS.cp.fDF.c <- c(estimSkSS.cp.fDF.c, LLSkSS.cp.fDF.c, LL.diffSkSS.cp.fDF.c)

TimeSkSS.cp.fDF.c <- system.time(
  while(abs(LL.diffSkSS.cp.fDF.c)>0.001){
    
    estSkSS.cp.fDF.c <- fit.hier.univ.SKnormSS.fDF(data = class.data.sk, parastart = estimSkSS.cp.fDF.c, parametrisation="cp", df=3, method="BFGS", hessian=FALSE, trace=2)
    LL.diffSkSS.cp.fDF.c <- estSkSS.cp.fDF.c$LL - LLSkSS.cp.fDF.c
    LLSkSS.cp.fDF.c <- estSkSS.cp.fDF.c$LL
    estimSkSS.cp.fDF.c <- estSkSS.cp.fDF.c$est
    
    resSkSS.cp.fDF.c <- rbind(resSkSS.cp.fDF.c, c(estimSkSS.cp.fDF.c, LLSkSS.cp.fDF.c, LL.diffSkSS.cp.fDF.c))
    
    estSkSS.cp.fDF.c <- fit.hier.univ.SKnormSS.fDF(data = class.data.sk, parastart = estimSkSS.cp.fDF.c, parametrisation="cp", df=3, method="Nelder-Mead", hessian=FALSE, trace=2)
    LL.diffSkSS.cp.fDF.c <- estSkSS.cp.fDF.c$LL - LLSkSS.cp.fDF.c
    LLSkSS.cp.fDF.c <- estSkSS.cp.fDF.c$LL
    estimSkSS.cp.fDF.c <- estSkSS.cp.fDF.c$est
    
    resSkSS.cp.fDF.c <- rbind(resSkSS.cp.fDF.c, c(estimSkSS.cp.fDF.c, LLSkSS.cp.fDF.c, LL.diffSkSS.cp.fDF.c))
    
  }
)


save(resBI.fDF, TimeBI.fDF, resBH.fDF, TimeBH.fDF, resS.fDF, TimeS.fDF, resIntS.fDF, TimeIntS.fDF, resS.fDF.c, TimeS.fDF.c, resSkSS.cp.fDF, TimeSkSS.cp.fDF, resIntSkSS.cp.fDF, TimeIntSkSS.cp.fDF, resSkSS.cp.fDF.c, TimeSkSS.cp.fDF.c, file="Results_LogIncome.RData" )


