library(gplots)
library(mvtnorm)
library(MASS)
library(sn)
library(MCMCpack)
library(microbenchmark) ## to record time in nano seconds
library(truncnorm)
library(xtable)

rm(list=ls())
source("functions_billard.R")
source("functions_uni_hist_N_SN.R")
source("functions_uni_int_N_SN.R")
source("functions_uni_N_SN.R")

### REQUIRES TO RUN LINE 1-82 OF LOG-INCOME.R

iter=1000

##################################################
########## Assuming Normal distribution  #########
##################################################

## first simulate 78 paramters where using MLE of Classical data assuming Normal distribution

mle_classical <-resS.fDF.c[which.max(resS.fDF.c[,77]), 1:76]
est = matrix(0, iter,76)

## Simulate the parameters at which the likelihoods are evaluated (1000 sets)

set.seed(42641)
for(j in 1:76){
  if( any(j == c(1:70)) ){
    if( j %% 2 == 1){ est[,j] = rnorm(iter,mle_classical[j], sd = 0.1 ) }
    if( j %% 2 == 0){ est[,j] = rtruncnorm(iter, 0, Inf, mle_classical[j], sd = 0.1 ) }
  }
  if( any(j == c(71:73))){ est[,j] = rnorm(iter,mle_classical[j], sd = 0.1 ) }
  if( any(j == c(74:76))){ est[,j] = rtruncnorm(iter, 0, Inf, mle_classical[j], sd = 0.1 ) }
}


## Evaluate the likelihoods

# Histogram Symbols
symbolic_norm <- vector(length=iter)
pb <-  txtProgressBar(min = 0, max = iter, initial = 0, style=3) #creates a prgress bar
for(i in 1:iter){
  setTxtProgressBar(pb, i)
  symbolic_norm[i] = microbenchmark(llik.hier.univ.hist.normS(symbol.data = symbols.hist, Theta = est[i,1:70] , varphi = c(est[i,71:74], 3, est[i,75:76])),times = 1)$time/(10^6)
  
}

# Classical 
classical_norm <- vector(length=iter)
pb <-  txtProgressBar(min = 0, max = iter, initial = 0, style=3) #creates a prgress bar
for(i in 1:iter){
  setTxtProgressBar(pb, i)
  classical_norm[i] = microbenchmark(llik.hier.univ.normS(data = class.data, Theta =est[i,1:70] , varphi = c(est[i,71:74], 3, est[i,75:76])),times = 1)$time/(10^6)
}

# Histogram symbols LRB
LRB <- vector(length=iter)
pb <-  txtProgressBar(min = 0, max = iter, initial = 0, style=3) #creates a prgress bar
for(i in 1:iter){
  setTxtProgressBar(pb, i)
  LRB[i] = microbenchmark(llik.billard2(symbol.data = symbols.B.hist,  param=c(est[i, 71:74], 3, est[i,75:76])), times = 1)$time/(10^6)
  
}


##########################################################################################
########## Assuming SN distribution with location scale skewness parameterisation ########
##########################################################################################

mle_sn_classical <- resSkSS.cp.fDF.c[which.max(resSkSS.cp.fDF.c[,114]), 1:113]
est_cp <- matrix(0, iter, 113)

## Generate a fake dataset with 1 million observation per symbol.
Time.Hist <- vector(length=35)
symbols.Hist <- symbols.B.Hist <- class.Data.sk <- list(length=35)
M.Hist <- S.Hist <- vector(length=35)

set.seed(14342)
for(i in 1:35){
  cp.par <- mle_sn_classical[((i-1)*3+1):(i*3)]
  cp.par <- c(cp.par[1], sqrt(cp.par[2]), cp.par[3]) # We modelled it as variance, not sd!

  n <- 1e+6
  tmp <- rsn(n=n, dp=cp2dp(cp.par, "SN"))
  
  # Classical data
  class.Data.sk[[i]] <- list(data = tmp, risk.lvl = i)

  # Histogram symbols  
  nclass <-  5 # Number of bins    
  x.breaks <- seq(from=min(tmp)-1e-10, to=max(tmp)+1e-10,  length=nclass+1 )    
  Time.Hist[i] <- microbenchmark(symbols.Hist[[i]] <- hist(tmp, breaks=x.breaks, plot=FALSE),times = 1)$time/(10^9)
  symbols.Hist[[i]]$risk.lvl <- i  

  # Histogram symbols (LRB)
  A <- head(symbols.Hist[[i]]$breaks,-1)
  B <- tail(symbols.Hist[[i]]$breaks,-1)
  M.Hist[i] <- sum((A+B) * symbols.Hist[[i]]$counts)  /(2*sum(symbols.Hist[[i]]$counts))
  S.Hist[i] <- sum(symbols.Hist[[i]]$counts * (A^2 + A*B + B^2) )/(3*n) - (M.Hist[i])^2
  symbols.B.Hist[[i]] <- list( val = c(M.Hist[i], S.Hist[i]), risk.lvl = i)
  
}

## Simulate 113 parameters where using MLE of Classical data assuming SN

set.seed(42641)
for(j in 1:113){
  if( any(j == c(1:105)) ){
    if( j %% 3 == 1){ est_cp[,j] = rnorm(iter,mle_sn_classical[j], sd = 0.1 ) }
    if( j %% 3 == 2){ est_cp[,j] = rtruncnorm(iter, 0, Inf, mle_sn_classical[j], sd = 0.1 ) }
    if( j %% 3 == 0){ est_cp[,j] = rnorm(iter,mle_sn_classical[j], sd = 0.1)} # Careful at bounds of the skewness parameter in cp parametrisation!
    # any(abs(est_cp[,seq(3,105,3)]) >= (4-pi)/2 * (2 / (pi-2))^(3/2))   
  }
  if( any(j == c(106:108, 112))){ est_cp[,j] = rnorm(iter,mle_sn_classical[j], sd = 0.1 ) }
  if( any(j == c(109:111, 113))){ est_cp[,j] = rtruncnorm(iter, 0, Inf, mle_sn_classical[j], sd = 0.1 ) }
}

## Evaluate the likelihoods

# Histogram Symbols
symbolic_sn_cp = matrix(0, iter, 2)
pb <-  txtProgressBar(min = 0, max = iter, initial = 0, style=3) #creates a prgress bar
for(i in 1:iter){
  setTxtProgressBar(pb, i)
  symbolic_sn_cp[i,1] = microbenchmark(llik.hier.univ.hist.SKnormSS(symbol.data = symbols.hist, Theta = est_cp[i,1:105], parametrisation="cp", varphi = c(est_cp[i,106:109], 3, est_cp[i,110:113])), times = 1)$time/(10^6)
  symbolic_sn_cp[i,2] = microbenchmark(llik.hier.univ.hist.SKnormSS(symbol.data = symbols.Hist, Theta = est_cp[i,1:105], parametrisation="cp", varphi = c(est_cp[i,106:109], 3, est_cp[i,110:113])), times = 1)$time/(10^6)
  
}

# Classical assuming SN (cp parametrisation)
classical_sn_cp <- matrix(0, iter, 2)
pb <-  txtProgressBar(min = 0, max = iter, initial = 0, style=3) #creates a prgress bar
for(i in 1:iter){
  setTxtProgressBar(pb, i)
  classical_sn_cp[i,1] = microbenchmark(llik.hier.univ.SKnormSS(data = class.data.sk, Theta = est_cp[i,1:105] , varphi = c(est_cp[i,106:109],3,est_cp[i,110:113])),times = 1)$time/(10^6)
  classical_sn_cp[i,2] = microbenchmark(llik.hier.univ.SKnormSS(data = class.Data.sk, Theta = est_cp[i,1:105] , varphi = c(est_cp[i,106:109],3,est_cp[i,110:113])),times = 1)$time/(10^6)
  
}

# Histogram symbols LRB
LRB_sn_cp <- matrix(0, iter, 2)
pb <-  txtProgressBar(min = 0, max = iter, initial = 0, style=3) #creates a prgress bar
for(i in 1:iter){
  setTxtProgressBar(pb, i)
  LRB_sn_cp[i,1] = microbenchmark(llik.billard2(symbol.data = symbols.B.hist, param = c(est_cp[i,106:109],3,est_cp[i,110:111])),times = 1)$time/(10^6)
  LRB_sn_cp[i,2] = microbenchmark(llik.billard2(symbol.data = symbols.B.Hist, param = c(est_cp[i,106:109],3,est_cp[i,110:111])),times = 1)$time/(10^6)
}

save(symbolic_norm, symbolic_sn_cp, classical_norm, classical_sn_cp, LRB_sn_cp, LRB, file="times_llik.RData")

results <- matrix(nrow = 3, ncol=5)
precision <- 3

classical.mean <- round( c(mean(classical_norm,na.rm=TRUE), apply(classical_sn_cp, 2, mean)), precision) 
classical.se <- round( c(sd(classical_norm,na.rm=TRUE), apply(classical_sn_cp, 2, sd) )/sqrt(iter), precision)

symb.mean <- round( c(mean(symbolic_norm,na.rm=TRUE), apply(symbolic_sn_cp, 2, mean)[c(1,2)]), precision) 
symb.se <- round( c(sd(symbolic_norm,na.rm=TRUE), apply(symbolic_sn_cp, 2, sd)[c(1,2)])/sqrt(iter), precision)

LRB.mean <- round( c(mean(LRB,na.rm=TRUE), apply(LRB_sn_cp, 2, mean) ), precision) 
LRB.se <- round( c(sd(LRB,na.rm=TRUE), apply(LRB_sn_cp, 2, sd) )/sqrt(iter), precision)

results[1,1] <- paste( format(classical.mean[1], nsmall=precision)," (", format(classical.se[1], nsmall=precision), ")",sep="")
results[1,2] <- paste( format(classical.mean[2], nsmall=precision)," (", format(classical.se[2], nsmall=precision), ")",sep="")
results[1,3] <- paste( format(classical.mean[3], nsmall=precision)," (", format(classical.se[3], nsmall=precision), ")",sep="")

results[2,1] <- paste( format(symb.mean[1], nsmall=precision)," (", format(symb.se[1], nsmall=precision), ")",sep="")
results[2,2] <- paste( format(symb.mean[2], nsmall=precision)," (", format(symb.se[2], nsmall=precision), ")",sep="")
results[2,3] <- paste( format(symb.mean[3], nsmall=precision)," (", format(symb.se[3], nsmall=precision), ")",sep="")

results[3,1] <- paste( format(LRB.mean[1], nsmall=precision)," (", format(LRB.se[1], nsmall=precision), ")",sep="")
results[3,2] <- paste( format(LRB.mean[2], nsmall=precision)," (", format(LRB.se[2], nsmall=precision), ")",sep="")
results[3,3] <- paste( format(LRB.mean[3], nsmall=precision)," (", format(LRB.se[3], nsmall=precision), ")",sep="")

results <- cbind(results[,1], rep(NA,3), results[,2], rep(NA,3), results[,3])
print(xtable(results, type = "latex", digits=precision))
